from qtpy.QtWidgets import QDialog

from ibeatles import load_ui
from ibeatles._version import __version__


class AboutLauncher(QDialog):
    def __init__(self, parent=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)

        self.ui = load_ui("about.ui", baseinstance=self)
        self.setWindowTitle("About")
        self.ui.application_version_label.setText(f"iBeatles: {__version__}")

        list_version = []
        try:
            import numpy
        except ImportError:
            list_version.append("numpy: unknown")
        else:
            list_version.append(f"numpy: {numpy.__version__}")

        try:
            import matplotlib
        except ImportError:
            list_version.append("matplotlib: unknown")
        else:
            list_version.append(f"matplotlib: {matplotlib.__version__}")

        try:
            import pyqtgraph
        except ImportError:
            list_version.append("pyqtgraph: unknown")
        else:
            list_version.append(f"pyqtgraph: {pyqtgraph.__version__}")

        try:
            import pandas
        except ImportError:
            list_version.append("pandas: unknown")
        else:
            list_version.append(f"pandas: {pandas.__version__}")

        try:
            import NeuNorm
        except ImportError:
            list_version.append("NeuNorm: unknown")
        else:
            list_version.append(f"NeuNorm: {NeuNorm.__version__}")

        try:
            import qtpy
        except ImportError:
            list_version.append("qtpy: unknown")
        else:
            list_version.append(f"qtpy: {qtpy.__version__}")

        try:
            import NeuNorm
        except ImportError:
            list_version.append("NeuNorm: unknown")
        else:
            list_version.append(f"NeuNorm: {NeuNorm.__version__}")

        try:
            import qtpy
        except ImportError:
            list_version.append("qtpy: unknown")
        else:
            list_version.append(f"qtpy: {qtpy.__version__}")

        formatted_list_version = "\n".join(list_version)
        self.ui.librairies_versions_textEdit.setText(formatted_list_version)



#!/usr/bin/env python
"""
Display class for Kropff fitting
"""

import numpy as np
import pyqtgraph as pg

from ibeatles.fitting.kropff.get import Get as GetKropff
from ibeatles.utilities.display import Display as UtilitiesDisplay


class Display:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def display_bragg_peak_threshold(self):
        # clear all previously display bragg peak threshold
        if self.parent.kropff_threshold_current_item is not None:
            self.parent.ui.kropff_fitting.removeItem(self.parent.kropff_threshold_current_item)
            self.parent.kropff_threshold_current_item = None

        # get list of row selected
        o_kropff = GetKropff(parent=self.parent)
        if o_kropff.kropff_row_selected():
            row_selected = str(o_kropff.kropff_row_selected()[0])
        else:
            return

        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        kropff_table_of_row_selected = kropff_table_dictionary[row_selected]

        # retrieve value of threshold range
        left = kropff_table_of_row_selected["bragg peak threshold"]["left"]
        right = kropff_table_of_row_selected["bragg peak threshold"]["right"]

        if left is None:
            return

        # display item and make it enabled or not according to is_manual mode or not
        lr = pg.LinearRegionItem(
            values=[left, right],
            orientation="vertical",
            brush=None,
            movable=True,
            bounds=None,
        )
        lr.setZValue(-10)
        lr.sigRegionChangeFinished.connect(self.parent.kropff_bragg_edge_threshold_changed)
        self.parent.ui.kropff_fitting.addItem(lr)
        self.parent.kropff_threshold_current_item = lr

        self.display_lambda_0()
        self.display_lambda_calculated()

    def update_fitting_parameters_matplotlib(self):
        o_get = GetKropff(parent=self.parent)
        matplotlib_ui = o_get.kropff_matplotlib_ui_selected()
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        fitting_parameter_to_plot = o_get.kropff_fitting_parameters_radioButton_selected()

        if not fitting_parameter_to_plot:
            return

        def use_error_bar_plot(array):
            for _value in array:
                if _value is None:
                    return False
            return True

        parameter_array = []
        parameter_error_array = []
        for _row in kropff_table_dictionary.keys():
            _value = kropff_table_dictionary[_row][fitting_parameter_to_plot]["val"]
            _error = kropff_table_dictionary[_row][fitting_parameter_to_plot]["err"]
            parameter_array.append(_value)
            parameter_error_array.append(_error)

        x_array = np.arange(len(parameter_array))
        matplotlib_ui.axes.cla()
        # if fit_region == 'bragg_peak':
        #     plot_ui.axes.set_yscale("log")

        if len(np.where(np.isnan(parameter_array))[0]) == len(parameter_array):
            # nothing to plot because all data points are NaN
            return

        if use_error_bar_plot(parameter_error_array):
            matplotlib_ui.axes.errorbar(
                x_array,
                parameter_array,
                parameter_error_array,
                marker="o",
                linestyle="",
                markeredgecolor="red",
                markerfacecolor="green",
                label="\u03bb_calculated",
                ecolor="red",
            )
        else:
            matplotlib_ui.axes.plot(
                x_array,
                parameter_array,
                marker="o",
                linestyle="",
                markeredgecolor="red",
                markerfacecolor="green",
                label="\u03bb_calculated",
            )

        matplotlib_ui.axes.set_xlabel("Row # (see Table tab)")
        matplotlib_ui.axes.set_ylabel(fitting_parameter_to_plot)

        # for the lambda_hkl, display the lambda_0 as a reference
        if fitting_parameter_to_plot == "lambda_hkl":
            lambda_0 = float(str(self.parent.ui.bragg_edge_calculated.text()))
            matplotlib_ui.axes.hlines(
                lambda_0,
                xmin=0,
                xmax=len(x_array),
                colors="b",
                linestyles="dotted",
                label="\u03bb\u2080",
            )
            matplotlib_ui.axes.legend(loc="upper center")

        matplotlib_ui.draw()

    def display_lambda_0(self):
        pyqtgraph_ui = self.parent.ui.kropff_fitting
        item = self.parent.lambda_0_item_in_kropff_fitting_plot
        lambda_position = float(str(self.parent.ui.bragg_edge_calculated.text()))

        o_utility_display = UtilitiesDisplay(ui=pyqtgraph_ui)
        new_item = o_utility_display.vertical_line(item=item, x_position=lambda_position)
        self.parent.lambda_0_item_in_kropff_fitting_plot = new_item

    def display_lambda_calculated(self):
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        o_kropff = GetKropff(parent=self.parent)
        row_selected = str(o_kropff.kropff_row_selected()[0])
        kropff_table_of_row_selected = kropff_table_dictionary[row_selected]
        lambda_value = kropff_table_of_row_selected["lambda_hkl"]["val"]

        # in bragg edge plot (top plot)
        item = self.parent.lambda_calculated_item_in_bragg_edge_plot
        o_utility_display = UtilitiesDisplay(ui=self.parent.bragg_edge_plot)
        new_item = o_utility_display.vertical_line(
            item=item,
            x_position=lambda_value,
            label="\u03bb_calculated",
            pen=pg.mkPen(color="g", width=1.5),
        )
        self.parent.lambda_calculated_item_in_bragg_edge_plot = new_item

        # in kropff fitting plot (bottom right)
        item = self.parent.lambda_calculated_item_in_kropff_fitting_plot
        o_utility_display = UtilitiesDisplay(ui=self.parent.ui.kropff_fitting)
        new_item = o_utility_display.vertical_line(
            item=item,
            x_position=lambda_value,
            label="\u03bb_calculated",
            pen=pg.mkPen(color="g", width=1.5),
        )
        self.parent.lambda_calculated_item_in_kropff_fitting_plot = new_item



#!/usr/bin/env python
"""
KropffAutomaticSettingsLauncher class for handling the automatic Bragg peak settings launcher.
"""

from qtpy.QtWidgets import QDialog

from ibeatles import DataType, load_ui
from ibeatles.fitting import FittingTabSelected
from ibeatles.fitting.kropff import KropffThresholdFinder
from ibeatles.fitting.kropff import SessionSubKeys as KropffSessionSubKeys


class KropffAutomaticSettingsLauncher(QDialog):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent
        super(QDialog, self).__init__(parent)
        self.ui = load_ui("ui_automatic_bragg_peak_settings.ui", baseinstance=self)
        self.init_widgets()

    def init_widgets(self):
        threshold_algo = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.automatic_bragg_peak_threshold_algorithm
        ]

        if threshold_algo == KropffThresholdFinder.sliding_average:
            self.ui.sliding_average_radioButton.setChecked(True)
        elif threshold_algo == KropffThresholdFinder.error_function:
            self.ui.error_function_radioButton.setChecked(True)
        elif threshold_algo == KropffThresholdFinder.change_point:
            self.ui.change_point_radioButton.setChecked(True)
        else:
            raise NotImplementedError("Algorithm not implemented!")

        # init threshold width
        fitting_width = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.automatic_fitting_threshold_width
        ]
        self.ui.kropff_threshold_width_slider.setValue(fitting_width)

    def save_algorithm_selected(self):
        if self.ui.sliding_average_radioButton.isChecked():
            algo_selected = KropffThresholdFinder.sliding_average
        elif self.ui.error_function_radioButton.isChecked():
            algo_selected = KropffThresholdFinder.error_function
        elif self.ui.change_point_radioButton.isChecked():
            algo_selected = KropffThresholdFinder.change_point
        else:
            raise NotImplementedError("Algorithm not implemented!")
        # self.parent.kropff_automatic_threshold_finder_algorithm = algo_selected
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.automatic_bragg_peak_threshold_algorithm
        ] = algo_selected

    def save_slider_value(self):
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.automatic_fitting_threshold_width
        ] = self.ui.kropff_threshold_width_slider.value()

    def slider_moved(self, _):
        self.slider_clicked()

    def slider_clicked(self):
        slider_value = self.ui.kropff_threshold_width_slider.value()
        self.ui.kropff_threshold_width_value.setText(str(slider_value))

    def slider_changed(self, _):
        self.slider_clicked()

    def ok_clicked(self):
        self.save_algorithm_selected()
        self.save_slider_value()
        self.close()



import numpy as np
from qtpy.QtWidgets import QMainWindow

from .. import load_ui


class InitializationSigmaAlpha(object):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

        init_sigma_alpha_window = InitializeWindow(parent=parent, grand_parent=grand_parent)
        init_sigma_alpha_window.show()


class InitializeWindow(QMainWindow):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent
        QMainWindow.__init__(self, parent=parent)
        self.ui = load_ui("ui_initSigmaAlpha.ui", baseinstance=self)
        self.init_widgets()

    def init_widgets(self):
        self.ui.alpha_error.setVisible(False)
        self.ui.sigma_error.setVisible(False)

    def ok_button_clicked(self):
        if self.variable_correctly_initialized():
            self.parent.sigma_alpha_initialized = True
            self.parent.initialize_all_parameters_step2()
            self.parent.is_ready_to_fit = True
            self.parent.check_state_of_step4_button()
            self.close()

    def cancel_button_clicked(self):
        self.parent.sigma_alpha_initialized = False
        self.close()

    def variable_correctly_initialized(self):
        _alpha = str(self.ui.alpha_lineEdit.text())
        _sigma = str(self.ui.sigma_lineEdit.text())

        _sigma_status = True
        try:
            _sigma = float(_sigma)
        except ValueError:
            _sigma = np.nan
            _sigma_status = False

        _alpha_status = True
        try:
            _alpha = float(_alpha)
        except ValueError:
            _alpha = np.nan
            _alpha_status = False

        initialization_table = self.parent.initialization_table
        initialization_table["sigma"] = _sigma
        initialization_table["alpha"] = _alpha
        self.grand_parent.initialization_table = initialization_table

        self.ui.sigma_error.setVisible(not _sigma_status)
        self.ui.alpha_error.setVisible(not _alpha_status)

        if _sigma_status and _alpha_status:
            return True
        else:
            return False



class CheckError(object):
    delay = 10000  # microseconds the message will be displayed in status bar

    def __init__(self, parent=None):
        self.parent = parent

        self.check_step1()

    def check_step1(self):
        self.check_file_errors()

    def check_file_errors(self):
        step = "step1"
        message = "Number of Data and OB files NOT compatible!"

        nbr_data_file = len(self.parent.list_files["sample"])

        if nbr_data_file == 0:
            return

        nbr_ob_file = len(self.parent.list_files["ob"])
        if nbr_ob_file == 0:
            return

        if not (nbr_data_file == nbr_ob_file):
            self.save_and_report_message(step=step, message=message, status=True)
            return

        self.save_and_report_message(step=step, message="", status=False)

    def save_and_report_message(self, **kwargs):
        self.save_report(**kwargs)
        self.report_message(step=kwargs["step"])

    def save_report(self, step="step1", message="", status=False):
        steps_error = self.parent.steps_error
        steps_error[step]["status"] = status
        steps_error[step]["message"] = message
        self.parent.steps_error = steps_error

    def report_message(self, step="step1"):
        steps_error = self.parent.steps_error
        if steps_error[step]["status"]:
            _message = steps_error[step]["message"]
            self.parent.ui.statusbar.showMessage(_message, self.delay)



#!/usr/bin/env python
"""
Bragg edge selection handler
"""

from qtpy.QtWidgets import QAbstractItemView

from ibeatles.utilities.array_utilities import find_nearest_index


class BraggEdgeSelectionHandler:
    def __init__(self, parent=None, data_type="sample"):
        self.parent = parent
        self.data_type = data_type

    def update_dropdown(self):
        lr = self.parent.list_bragg_edge_selection_id[self.data_type]
        x_axis = self.parent.current_bragg_edge_x_axis[self.data_type]

        selection = list(lr.getRegion())

        left_index = find_nearest_index(array=x_axis, value=selection[0])
        right_index = find_nearest_index(array=x_axis, value=selection[1])

        list_selected = range(left_index, right_index + 1)

        if self.data_type == "sample":
            _ui_list = self.parent.ui.list_sample
        elif self.data_type == "ob":
            _ui_list = self.parent.ui.list_open_beam
        else:
            _ui_list = self.parent.ui.list_normalized

        first_item = True

        nbr_item = _ui_list.count()
        for _row in range(nbr_item):
            item = _ui_list.item(_row)
            if _row in list_selected:
                if first_item:
                    first_item = False
                    _ui_list.scrollToItem(item, QAbstractItemView.PositionAtTop)
                select_flag = True
            else:
                select_flag = False
            item.setSelected(select_flag)



#!/usr/bin/env python
"""
GUI handler for the step3
"""

from ibeatles import DataType
from ibeatles.utilities.retrieve_data_infos import (
    RetrieveGeneralDataInfos,
    RetrieveGeneralFileInfos,
)


class Step3GuiHandler:
    def __init__(self, parent=None):
        self.parent = parent

    def load_normalized_changed(self, tab_index=0):
        if tab_index == 0:
            data_preview_box_label = "Sample Image Preview"
            o_general_infos = RetrieveGeneralFileInfos(parent=self.parent, data_type="sample")
            o_selected_infos = RetrieveGeneralDataInfos(parent=self.parent, data_type="sample")
        else:
            data_preview_box_label = "Open Beam Image Preview"
            o_general_infos = RetrieveGeneralFileInfos(parent=self.parent, data_type="ob")
            o_selected_infos = RetrieveGeneralDataInfos(parent=self.parent, data_type="ob")

        self.parent.ui.data_preview_box.setTitle(data_preview_box_label)
        o_general_infos.update()
        o_selected_infos.update()

    def select_normalized_row(self, row=0):
        self.parent.ui.list_normalized.setCurrentRow(row)

    def check_time_spectra_widgets(self):
        time_spectra_data = self.parent.data_metadata["time_spectra"]["normalized_folder"]
        if self.parent.ui.material_display_checkbox.isChecked():
            if len(time_spectra_data) == 0:
                _display_error_label = True
            else:
                _display_error_label = False
        else:
            _display_error_label = False

        self.parent.ui.display_warning.setVisible(_display_error_label)

    def check_widgets(self):
        if len(self.parent.list_files[DataType.normalized]) == 0:
            status = False
        else:
            status = True

        self.parent.ui.actionRotate_Images.setEnabled(status)



#!/usr/bin/env python
"""
Filling table handler
"""

import numpy as np
from qtpy import QtCore, QtGui
from qtpy.QtWidgets import QCheckBox, QTableWidgetItem

from ibeatles.fitting import FittingKeys
from ibeatles.fitting.kropff import SessionSubKeys as KropffsessionSubKeys
from ibeatles.utilities.table_handler import TableHandler


class FillingTableHandler:
    table_dictionary: dict = {}
    advanced_mode_flag = True

    def __init__(self, grand_parent=None, parent=None):
        self.grand_parent = grand_parent
        self.parent = parent

    def set_mode(self, advanced_mode=True):
        self.advanced_mode_flag = advanced_mode
        list_header_table_advanced_columns = [10, 11]
        list_parent_advanced_columns = [15, 16, 17, 18]

        self.parent.ui.header_table.horizontalHeader().blockSignals(True)
        self.parent.ui.value_table.horizontalHeader().blockSignals(True)

        # hide column a5 and a6
        for _col_index in list_header_table_advanced_columns:
            self.parent.ui.header_table.setColumnHidden(_col_index, not advanced_mode)
        for _col_index in list_parent_advanced_columns:
            self.parent.ui.value_table.setColumnHidden(_col_index, not advanced_mode)

        self.parent.ui.header_table.horizontalHeader().blockSignals(False)
        self.parent.ui.value_table.horizontalHeader().blockSignals(False)

        # repopulate table
        self.fill_table()

    def get_row_to_show_state(self):
        """
        return 'all', 'active' or 'lock'
        """
        if self.parent.ui.show_all_bins.isChecked():
            return "all"
        elif self.parent.ui.show_only_active_bins.isChecked():
            return "active"
        else:
            return "lock"

    def fill_table(self):
        self.fill_march_table()
        self.fill_kropff_table()

    def fill_kropff_table(self):
        self.fill_kropff_high_tof_table()
        self.fill_kropff_low_tof_table()
        self.fill_kropff_bragg_peak_table()

    def fill_kropff_high_tof_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)
        o_table.block_signals(True)

        o_table.remove_all_rows()
        table_dictionary = self.grand_parent.kropff_table_dictionary
        nbr_row = len(table_dictionary)

        for _index in np.arange(nbr_row):
            _str_index = str(_index)

            _entry = table_dictionary[_str_index]

            o_table.insert_empty_row(row=_index)

            o_table.insert_item(
                row=_index,
                column=0,
                value=_entry[FittingKeys.row_index] + 1,
                editable=False,
                align_center=True,
            )
            o_table.insert_item(
                row=_index,
                column=1,
                value=_entry[FittingKeys.column_index] + 1,
                editable=False,
                align_center=True,
            )

            # from column 2 to 5
            list_value = [
                _entry[KropffsessionSubKeys.a0]["val"],
                _entry[KropffsessionSubKeys.b0]["val"],
                _entry[KropffsessionSubKeys.a0]["err"],
                _entry[KropffsessionSubKeys.b0]["err"],
            ]

            for _local_index, _value in enumerate(list_value):
                o_table.insert_item(
                    row=_index,
                    column=_local_index + 2,
                    value=_value,
                    format_str="{:.4f}",
                    editable=False,
                    align_center=True,
                )

        # select first row
        o_table.block_signals(True)
        o_table.select_row(0)

        o_table.block_signals(False)

    def fill_kropff_low_tof_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.low_lda_tableWidget)
        o_table.remove_all_rows()
        table_dictionary = self.grand_parent.kropff_table_dictionary
        nbr_row = len(table_dictionary)
        o_table.block_signals(True)

        for _index in np.arange(nbr_row):
            _str_index = str(_index)
            _entry = table_dictionary[_str_index]

            o_table.insert_empty_row(row=_index)

            o_table.insert_item(
                row=_index,
                column=0,
                value=_entry[FittingKeys.row_index] + 1,
                editable=False,
                align_center=True,
            )
            o_table.insert_item(
                row=_index,
                column=1,
                value=_entry[FittingKeys.column_index] + 1,
                editable=False,
                align_center=True,
            )

            # from column 2 to 5
            list_value = [
                _entry[KropffsessionSubKeys.ahkl]["val"],
                _entry[KropffsessionSubKeys.bhkl]["val"],
                _entry[KropffsessionSubKeys.ahkl]["err"],
                _entry[KropffsessionSubKeys.bhkl]["err"],
            ]
            for _local_index, _value in enumerate(list_value):
                o_table.insert_item(
                    row=_index,
                    column=_local_index + 2,
                    value=_value,
                    format_str="{:.4f}",
                    editable=False,
                    align_center=True,
                )

        # select first row
        o_table.select_row(0)

        o_table.block_signals(False)

    def fill_kropff_bragg_peak_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)
        o_table.remove_all_rows()
        table_dictionary = self.grand_parent.kropff_table_dictionary
        nbr_row = len(table_dictionary)
        o_table.block_signals(True)

        for _index in np.arange(nbr_row):
            _str_index = str(_index)
            _entry = table_dictionary[_str_index]

            o_table.insert_empty_row(row=_index)

            o_table.insert_item(
                row=_index,
                column=0,
                value=_entry[FittingKeys.row_index] + 1,
                editable=False,
                align_center=True,
            )
            o_table.insert_item(
                row=_index,
                column=1,
                value=_entry[FittingKeys.column_index] + 1,
                editable=False,
                align_center=True,
            )

            # from column 2 to 7
            list_value = [
                _entry[KropffsessionSubKeys.lambda_hkl]["val"],
                _entry[KropffsessionSubKeys.tau]["val"],
                _entry[KropffsessionSubKeys.sigma]["val"],
                _entry[KropffsessionSubKeys.lambda_hkl]["err"],
                _entry[KropffsessionSubKeys.tau]["err"],
                _entry[KropffsessionSubKeys.sigma]["err"],
            ]

            for _local_index, _value in enumerate(list_value):
                format_str = ""
                if _value:
                    number_of_digits = self.parent.ui.kropff_bragg_peak_number_of_digits_spinBox.value()
                    format_str = "{:." + str(number_of_digits) + "f}"
                o_table.insert_item(
                    row=_index,
                    column=_local_index + 2,
                    value=_value,
                    format_str=format_str,
                    editable=False,
                    align_center=True,
                )

        # select first row
        o_table.select_row(0)

        o_table.block_signals(False)

    def fill_march_table(self):
        self.clear_table_ui()
        table_dictionary = self.grand_parent.march_table_dictionary

        row_to_show_state = self.get_row_to_show_state()
        nbr_row = len(table_dictionary)

        self.parent.ui.value_table.blockSignals(True)

        for _index in np.arange(nbr_row):
            _str_index = str(_index)
            _entry = table_dictionary[_str_index]

            # add new row
            self.parent.ui.value_table.insertRow(_index)

            # row and column index (columns 0 and 1)
            self.set_item(
                table_ui=self.parent.ui.value_table,
                row=_index,
                col=0,
                value=_entry[FittingKeys.row_index] + 1,
            )  # +1 because table starts indexing at 1

            self.set_item(
                table_ui=self.parent.ui.value_table,
                row=_index,
                col=1,
                value=_entry[FittingKeys.column_index] + 1,
            )  # +1 because table starts indexing at 1

            # add lock button in first cell (column: 2)
            _lock_button = QCheckBox()
            _is_lock = _entry[FittingKeys.lock]

            _lock_button.setChecked(_is_lock)
            _lock_button.stateChanged.connect(
                lambda state=0, row=_index: self.parent.lock_button_state_changed(state, row)
            )

            self.parent.ui.value_table.setCellWidget(_index, 2, _lock_button)

            # add active button in second cell (column: 3)
            _active_button = QCheckBox()
            _is_active = _entry[FittingKeys.active]

            _active_button.setChecked(_is_active)
            _active_button.stateChanged.connect(
                lambda state=0, row=_index: self.parent.active_button_state_changed(state, row)
            )

            self.parent.ui.value_table.setCellWidget(_index, 3, _active_button)

            # bin # (column: 1)
            # _bin_number = QTableWidgetItem("{:02}".format(_index))
            # grand_parent_ui.setItem(_index, 1, _bin_number)

            # from column 2 -> nbr_column
            list_value = [
                _entry["fitting_confidence"],
                _entry["d_spacing"]["val"],
                _entry["d_spacing"]["err"],
                _entry["sigma"]["val"],
                _entry["sigma"]["err"],
                _entry["alpha"]["val"],
                _entry["alpha"]["err"],
                _entry["a1"]["val"],
                _entry["a1"]["err"],
                _entry["a2"]["val"],
                _entry["a2"]["err"],
                _entry["a5"]["val"],
                _entry["a5"]["err"],
                _entry["a6"]["val"],
                _entry["a6"]["err"],
            ]

            list_fixed_flag = [
                False,
                _entry["d_spacing"]["fixed"],
                _entry["d_spacing"]["fixed"],
                _entry["sigma"]["fixed"],
                _entry["sigma"]["fixed"],
                _entry["alpha"]["fixed"],
                _entry["alpha"]["fixed"],
                _entry["a1"]["fixed"],
                _entry["a1"]["fixed"],
                _entry["a2"]["fixed"],
                _entry["a2"]["fixed"],
                _entry["a5"]["fixed"],
                _entry["a5"]["fixed"],
                _entry["a6"]["fixed"],
                _entry["a6"]["fixed"],
            ]

            for _local_index, _value in enumerate(list_value):
                self.set_item(
                    table_ui=self.parent.ui.value_table,
                    row=_index,
                    col=_local_index + 4,
                    value=_value,
                    fixed_flag=list_fixed_flag[_local_index],
                )

            if row_to_show_state == "active":
                if not _is_active:
                    self.parent.ui.value_table.hideRow(_index)
            elif row_to_show_state == "lock":
                if not _is_lock:
                    self.parent.ui.value_table.hideRow(_index)

        self.parent.ui.value_table.blockSignals(False)

    def set_item(self, table_ui=None, row=0, col=0, value="", fixed_flag=False):
        item = QTableWidgetItem(str(value))
        if fixed_flag:
            item.setTextColor(QtGui.QColor(255, 0, 0, alpha=255))
        item.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)

        table_ui.setItem(row, col, item)

    def clear_table_ui(self):
        self.parent.ui.blockSignals(True)
        nbr_row = self.parent.ui.value_table.rowCount()
        for _row in np.arange(nbr_row):
            self.parent.ui.value_table.removeRow(0)
        self.parent.ui.value_table.blockSignals(False)

    def clear_table(self):
        self.unselect_full_table()
        self.parent.ui.value_table.blockSignals(True)
        nbr_row = self.parent.ui.value_table.rowCount()
        for _row in np.arange(nbr_row):
            self.parent.ui.value_table.removeRow(0)
        self.parent.ui.value_table.blockSignals(False)

        self.parent.selection_in_grand_parent_changed()



#!/usr/bin/env python
"""Mathematical models used in Kropff fitting"""

import numpy as np
from scipy import special


def kropff_high_lambda_transmission(wavelength: np.ndarray, a0: float, b0: float) -> np.ndarray:
    """
    Calculate the high-wavelength side transmission function.

    Parameters
    ----------
    wavelength : np.ndarray
        The wavelength values to calculate the function.
    a0 : float
        The a0 parameter.
    b0 : float
        The b0 parameter.

    Returns
    -------
    np.ndarray
        The high-wavelength side transmission function.
    """
    return np.exp(-(a0 + b0 * wavelength))


def kropff_low_lambda_transmission(
    wavelength: np.ndarray, a0: float, b0: float, a_hkl: float, b_hkl: float
) -> np.ndarray:
    """
    Calculate the low-wavelength side transmission function.

    Parameters
    ----------
    wavelength : np.ndarray
        The wavelength values to calculate the function.
    a0 : float
        The a0 parameter.
    b0 : float
        The b0 parameter.
    a_hkl : float
        The a_hkl parameter.
    b_hkl : float
        The b_hkl parameter.

    Returns
    -------
    np.ndarray
        The low-wavelength side transmission function.
    """
    return np.exp(-(a0 + b0 * wavelength) - (a_hkl + b_hkl * wavelength))


def bragg_edge_function(wavelength: np.ndarray, bragg_edge_wavelength: float, sigma: float, tau: float) -> np.ndarray:
    """
    Calculate the Bragg edge function.

    Parameters
    ----------
    wavelength : np.ndarray
        The wavelength values to calculate the function.
    bragg_edge_wavelength : float
        The wavelength of the Bragg edge.
    sigma : float
        The sigma parameter (related to symmetric broadening).
    tau : float
        The tau parameter (related to asymmetric broadening).

    Returns
    -------
    np.ndarray
        The Bragg edge function.
    """
    t = wavelength - bragg_edge_wavelength
    y = sigma / tau

    term1 = special.erfc(-t / (np.sqrt(2) * sigma))
    term2 = np.exp(-(t / tau) + (y**2 / 2))
    term3 = special.erfc(-t / (np.sqrt(2) * sigma) + y)

    return 0.5 * (term1 - term2 * term3)


def kropff_transmission_model(
    wavelength: np.ndarray,
    a0: float,
    b0: float,
    a_hkl: float,
    b_hkl: float,
    bragg_edge_wavelength: float,
    sigma: float,
    tau: float,
) -> np.ndarray:
    """
    Calculate the full Kropff transmission model.

    Parameters
    ----------
    wavelength : np.ndarray
        The wavelength values to calculate the function.
    a0 : float
        The a0 parameter.
    b0 : float
        The b0 parameter.
    a_hkl : float
        The a_hkl parameter.
    b_hkl : float
        The b_hkl parameter.
    bragg_edge_wavelength : float
        The wavelength of the Bragg edge.
    sigma : float
        The sigma parameter (related to symmetric broadening).
    tau : float
        The tau parameter (related to asymmetric broadening).

    Returns
    -------
    np.ndarray
        The full Kropff transmission model.
    """
    high_lambda = kropff_high_lambda_transmission(wavelength, a0, b0)
    low_lambda = kropff_low_lambda_transmission(wavelength, a0, b0, a_hkl, b_hkl)
    edge_function = bragg_edge_function(wavelength, bragg_edge_wavelength, sigma, tau)

    return low_lambda + (high_lambda - low_lambda) * edge_function



#!/usr/bin/env python
"""
Initialization (step 2)
"""

import pyqtgraph as pg
from pyqtgraph.dockarea import Dock, DockArea
from qtpy.QtWidgets import (
    QHBoxLayout,
    QRadioButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from ibeatles import DEFAULT_NORMALIZATION_ROI, DataType
from ibeatles.step2 import roi_label_color
from ibeatles.utilities.colors import pen_color


class Initialization:
    col_width = [70, 50, 50, 50, 50]

    def __init__(self, parent=None):
        self.parent = parent

    def all(self):
        self.table()
        self.pyqtgraph()
        self.splitter()

    def splitter(self):
        self.parent.step2_ui["area"].setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/vertical_splitter_handle.png");
                                     }
                                     """)

    def table(self):
        for _index, _width in enumerate(self.col_width):
            self.parent.ui.normalization_tableWidget.setColumnWidth(_index, _width)

    def pyqtgraph(self):
        area = DockArea()
        area.setVisible(False)
        d1 = Dock("Sample", size=(200, 300))
        d2 = Dock("STEP1: Background normalization", size=(200, 100))
        # d3 = Dock("STEP2: Working Range Selection", size=(200, 100))

        area.addDock(d1, "top")
        # area.addDock(d3, 'bottom')
        area.addDock(d2, "bottom")
        # area.moveDock(d2, 'above', d3)

        # preview_widget = pg.GraphicsLayoutWidget()
        pg.setConfigOptions(antialias=True)

        vertical_layout = QVBoxLayout()
        # preview_widget.setLayout(vertical_layout)

        # image view
        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()

        # vertical_layout.addWidget(image_view)
        # top_right_widget = QWidget()
        d1.addWidget(image_view)

        # bragg edge plot
        bragg_edge_plot = pg.PlotWidget()
        bragg_edge_plot.plot()

        # # bragg_edge_plot.setLabel("top", "")
        # p1 = bragg_edge_plot.plotItem
        # p1.layout.removeItem(p1.getAxis('top'))
        # caxis = CustomAxis(orientation='top', parent=p1)
        # caxis.setLabel('')
        # caxis.linkToView(p1.vb)
        # p1.layout.addItem(caxis, 1, 1)

        # add file_index, TOF, Lambda x-axis buttons
        hori_layout = QHBoxLayout()
        button_widgets = QWidget()
        button_widgets.setLayout(hori_layout)

        # file index
        file_index_button = QRadioButton()
        file_index_button.setText("File Index")
        file_index_button.setChecked(True)
        # self.parent.connect(file_index_button, QtCore.SIGNAL("clicked()"),
        #                     self.parent.step2_file_index_radio_button_clicked)
        file_index_button.pressed.connect(self.parent.step2_file_index_radio_button_clicked)

        # tof
        tof_button = QRadioButton()
        tof_button.setText("TOF")
        # self.parent.connect(tof_button, QtCore.SIGNAL("clicked()"),
        #                     self.parent.step2_tof_radio_button_clicked)
        tof_button.pressed.connect(self.parent.step2_tof_radio_button_clicked)

        # lambda
        lambda_button = QRadioButton()
        lambda_button.setText("\u03bb")
        # self.parent.connect(lambda_button, QtCore.SIGNAL("clicked()"),
        #                     self.parent.step2_lambda_radio_button_clicked)
        lambda_button.pressed.connect(self.parent.step2_lambda_radio_button_clicked)

        spacer1 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hori_layout.addItem(spacer1)
        hori_layout.addWidget(file_index_button)
        hori_layout.addWidget(tof_button)
        hori_layout.addWidget(lambda_button)
        spacer2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hori_layout.addItem(spacer2)

        d2.addWidget(bragg_edge_plot)
        d2.addWidget(button_widgets)

        vertical_layout.addWidget(area)
        self.parent.ui.normalization_left_widget.setLayout(vertical_layout)

        self.parent.step2_ui["area"] = area
        self.parent.step2_ui["image_view"] = image_view
        self.parent.step2_ui["bragg_edge_plot"] = bragg_edge_plot
        # self.parent.step2_ui['normalized_profile_plot'] = normalized_profile_plot
        # self.parent.step2_ui['caxis'] = caxis
        self.parent.step2_ui["xaxis_file_index"] = file_index_button
        self.parent.step2_ui["xaxis_lambda"] = lambda_button
        self.parent.step2_ui["xaxis_tof"] = tof_button

        self.parent.xaxis_button_ui["normalization"]["tof"] = tof_button
        self.parent.xaxis_button_ui["normalization"]["file_index"] = file_index_button
        self.parent.xaxis_button_ui["normalization"]["lambda"] = lambda_button

    def roi(self):
        image_view = self.parent.step2_ui["image_view"]

        if self.parent.list_roi[DataType.normalization]:
            for _roi_id in self.parent.list_roi_id[DataType.normalization]:
                image_view.removeItem(_roi_id)
            for _label_roi_id in self.parent.list_label_roi_id[DataType.normalization]:
                image_view.removeItem(_label_roi_id)

            list_roi = self.parent.list_roi[DataType.normalization]
            list_roi_id = []
            list_label_roi_id = []

            for _roi in list_roi:
                [is_visible, x0, y0, width, height, region_type] = _roi
                x0 = int(x0)
                y0 = int(y0)
                width = int(width)
                height = int(height)

                roi = pg.ROI([x0, y0], [width, height], pen=pen_color["0"])
                roi.addScaleHandle([1, 1], [0, 0])
                image_view.addItem(roi)
                roi.sigRegionChanged.connect(self.parent.normalization_manual_roi_changed)

                label_roi = pg.TextItem(
                    html=f'<div style="text-align: center"><span style="color: '
                    f'{roi_label_color[str(region_type)]};">' + region_type + "</span></div>",
                    anchor=(-0.3, 1.3),
                    border="w",
                    fill=(0, 0, 255, 50),
                )
                label_roi.setPos(x0, y0)
                image_view.addItem(label_roi)

                label_roi.setVisible(is_visible)
                roi.setVisible(is_visible)

                list_roi_id.append(roi)
                list_label_roi_id.append(label_roi)

            self.parent.list_roi_id[DataType.normalization] = list_roi_id
            self.parent.list_label_roi_id[DataType.normalization] = list_label_roi_id

        else:
            self.parent.list_roi[DataType.normalization] = DEFAULT_NORMALIZATION_ROI
            [_, x0, y0, width, height, region_type] = DEFAULT_NORMALIZATION_ROI
            x0 = int(x0)
            y0 = int(y0)
            width = int(width)
            height = int(height)

            roi = pg.ROI([x0, y0], [width, height], pen=pen_color["0"])
            roi.addScaleHandle([1, 1], [0, 0])
            image_view.addItem(roi)
            roi.sigRegionChanged.connect(self.parent.normalization_manual_roi_changed)

            label_roi = pg.TextItem(
                html=f'<div style="text-align: center"><span style="color: '
                f'{roi_label_color[region_type]};">' + region_type + "</span></div>",
                anchor=(-0.3, 1.3),
                border="w",
                fill=(0, 0, 255, 50),
            )
            label_roi.setPos(x0, y0)
            image_view.addItem(label_roi)
            self.parent.list_roi_id[DataType.normalization] = [roi]
            self.parent.list_label_roi_id[DataType.normalization] = [label_roi]



import json
import os


class ConfigHandler:
    def __init__(self, parent=None):
        self.parent = parent

    def load(self):
        config_file_name = os.path.join(os.path.dirname(__file__), "config.json")
        with open(config_file_name) as f:
            config = json.load(f)
        self.parent.config = config

        for _homepath in config["homepath"]:
            if os.path.exists(_homepath):
                self.parent.homepath = _homepath



#!/usr/bin/env python
"""
Time spectra module
"""

import logging
from pathlib import Path

import numpy as np

from ibeatles.utilities.file_handler import FileHandler


def export_time_stamp_file(counts_array=None, tof_array=None, file_index_array=None, export_folder=None):
    """
    modify the time_spectra file to mirror the new bins
    :param counts_array: total counts of the given bin
           tof_array: original tof_array (coming from original spectra file)
           file_index_array: bin index
           export_folder: where to export that new time stamp file
    :return: None
    """
    time_spectra_file_name = str(Path(export_folder) / "image_Spectra.txt")
    new_tof_array = []
    for _list_files_index in file_index_array:
        list_tof = [tof_array[_index] for _index in _list_files_index]
        new_tof_array.append(np.mean(list_tof))

    file_content = ["shutter_time,counts"]
    for _tof, _counts in zip(new_tof_array, counts_array):
        file_content.append(f"{_tof},{_counts}")

    FileHandler.make_ascii_file(data=file_content, output_file_name=time_spectra_file_name)

    logging.info(f"Exported the new time spectra file: {time_spectra_file_name} ... Done!")



import logging

from qtpy.QtWidgets import QMessageBox

from .. import DataType
from ..utilities.status_message_config import StatusMessageStatus, show_status_message


class EventHandler:
    def __init__(self, parent=None, data_type="sample"):
        self.parent = parent
        self.data_type = data_type

    def _display_status_message_warning(self, message=""):
        show_status_message(
            parent=self.parent,
            status=StatusMessageStatus.warning,
            message=message,
            duration_s=5,
        )

    def is_step_selected_allowed(self, step_index_requested=0):
        """0: load data
        1: normalization
        2: normalized
        3: bin
        4: fit
        5: strain mapping
        6: rotation
        """

        # load tab
        # validate all the time
        if step_index_requested == 0:
            return True

        # normalization
        # validate only if data loaded
        if step_index_requested == 1:
            if len(self.parent.data_metadata[DataType.sample]["data"]) == 0:
                message = "Please load some sample data!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            return True

        # normalized
        # validate all the time
        if step_index_requested == 2:
            return True

        # bin
        # validate only if normalized data loaded
        if step_index_requested == 3:
            if len(self.parent.data_metadata[DataType.normalized]["data"]) == 0:
                message = "Please load some normalized data!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            return True

        # fitting
        # validate if there is a bin region selected
        if step_index_requested == 4:
            if len(self.parent.data_metadata[DataType.normalized]["data"]) == 0:
                message = "Please load some normalized data!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            if not self.parent.there_is_a_roi:
                message = "Please select a region to bin first (step binning)!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            return True

        # strain mapping
        # validate if fitting has been performed
        if step_index_requested == 5:
            if (self.parent.march_table_dictionary == {}) and (self.parent.kropff_table_dictionary == {}):
                message = "Please fit the data to be able to visualize the strain mapping!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            return True

        # rotation
        # validate if normalized data loaded
        if step_index_requested == 6:
            if len(self.parent.data_metadata[DataType.normalized]["data"]) == 0:
                message = "Please load some normalized data!"
                self._display_status_message_warning(message=message)
                self._display_message_box(message=message)
                EventHandler._update_logging(step_requested=step_index_requested, message=message)
                return False
            return True

        return True

    def _display_message_box(self, message=""):
        dlg = QMessageBox(self.parent)
        dlg.setWindowTitle("Unable to start this step!")
        dlg.setText(message)
        dlg.setStandardButtons(QMessageBox.Ok)
        dlg.setIcon(QMessageBox.Warning)
        button = dlg.exec()

        if button == QMessageBox.Ok:
            dlg.close()

    @staticmethod
    def _update_logging(step_requested=-1, message=""):
        logging.info(f"Error requesting step #{step_requested}")
        logging.info(message)



#!/usr/bin/env python
"""Normalization functions for the TOF imaging data."""

from typing import Tuple, Union, overload

import numpy as np
import scipy.ndimage

from ibeatles.core.config import KernelSize, KernelType, MovingAverage


@overload
def moving_average(
    data: np.ndarray,
    kernel_type: str,
    kernel: Union[Tuple[int, int], Tuple[int, int, int]],
) -> np.ndarray:
    """
    Apply moving average filter to the input data.

    Parameters
    ----------
    data : np.ndarray
        Input data. Must be a 2D image or 3D volume (TOF stack of 2D images).
    kernel_type : str
        Type of kernel to use for the filter. Must be either 'Box' or 'Gaussian'.
    kernel : Union[Tuple[int, int], Tuple[int, int, int]]
        Size of the kernel. For 2D data, use (y, x). For 3D data, use (y, x, lambda).

    Returns
    -------
    np.ndarray
        Filtered data with the same shape as the input.

    Raises
    ------
    ValueError
        If input data, kernel_type, or kernel is invalid.
    """
    pass


@overload
def moving_average(data: np.ndarray, config: MovingAverage) -> np.ndarray:
    """
    Apply moving average filter to the input data using a MovingAverage configuration.

    Parameters
    ----------
    data : np.ndarray
        Input data. Must be a 2D image or 3D volume (TOF stack of 2D images).
    config : MovingAverage
        Configuration for the moving average filter.

    Returns
    -------
    np.ndarray
        Filtered data with the same shape as the input.

    Raises
    ------
    ValueError
        If input data or configuration is invalid.
    """
    pass


def moving_average(
    data: np.ndarray,
    arg2: Union[str, MovingAverage],
    arg3: Union[Tuple[int, ...], None] = None,
) -> np.ndarray:
    """
    Apply moving average filter to the input data.

    This function supports two calling conventions:
    1. moving_average(data, kernel_type, kernel)
    2. moving_average(data, config)

    Parameters
    ----------
    data : np.ndarray
        Input data. Must be a 2D image or 3D volume (TOF stack of 2D images).
    arg2 : Union[str, MovingAverage]
        Either a string specifying the kernel type ('Box' or 'Gaussian') or a MovingAverage configuration object.
    arg3 : Union[Tuple[int, ...], None], optional
        If arg2 is a string, this should be a tuple specifying the kernel size.

    Returns
    -------
    np.ndarray
        Filtered data with the same shape as the input.

    Raises
    ------
    ValueError
        If input data, kernel type, kernel size, or configuration is invalid.
    """
    if isinstance(arg2, str):
        kernel_type = arg2
        kernel = arg3
        if kernel is None:
            raise ValueError("Kernel size must be provided when specifying kernel type as a string.")

        # Convert tuple to KernelSize
        if len(kernel) == 2:
            kernel_size = KernelSize(y=kernel[0], x=kernel[1])
            dimension = "2D"
        elif len(kernel) == 3:
            kernel_size = KernelSize(y=kernel[0], x=kernel[1], lambda_=kernel[2])
            dimension = "3D"
        else:
            raise ValueError("Kernel must be 2D or 3D.")

        config = MovingAverage(
            active=True,
            dimension=dimension,
            size=kernel_size,
            type=KernelType(kernel_type),
        )
    elif isinstance(arg2, MovingAverage):
        config = arg2
    else:
        raise ValueError("Invalid argument type for kernel_type or config.")

    if not config.active:
        return data

    if data.ndim not in (2, 3):
        raise ValueError("Data must be 2D image or 3D volume (TOF stack of 2D images).")

    kernel = _get_kernel_from_config(config)

    if data.ndim == 2 and len(kernel) == 3:
        raise ValueError("Cannot apply 3D kernel to 2D data.")

    if data.ndim == 3 and len(kernel) == 2:
        return _apply_2d_kernel_to_3d_data(data, config)

    if config.type == KernelType.box:
        return _apply_box_filter(data, kernel)
    elif config.type == KernelType.gaussian:
        return scipy.ndimage.gaussian_filter(data, kernel)
    else:
        raise ValueError(f"Unsupported kernel type: {config.type}")


def _get_kernel_from_config(config: MovingAverage) -> Tuple[int, ...]:
    """
    Extract kernel size from MovingAverage configuration.

    Parameters
    ----------
    config : MovingAverage
        The moving average configuration.

    Returns
    -------
    Tuple[int, ...]
        A tuple representing the kernel size.
    """
    kernel = (config.size.y, config.size.x)
    if config.dimension == "3D":
        kernel += (config.size.lambda_,)
    return kernel


def _apply_2d_kernel_to_3d_data(data: np.ndarray, config: MovingAverage) -> np.ndarray:
    output = np.zeros_like(data)
    for i in range(data.shape[2]):
        output[:, :, i] = moving_average(data[:, :, i], config)
    return output


def _apply_box_filter(data: np.ndarray, kernel: Tuple[int, ...]) -> np.ndarray:
    kernel_array = np.ones(kernel)
    kernel_array /= kernel_array.sum()
    return scipy.ndimage.convolve(data, kernel_array)



#!/usr/bin/env python
"""
TOF binning launcher
"""

import logging
import warnings

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow

from ibeatles import load_ui
from ibeatles.tools.tof_bin import BinAutoMode, BinMode, session
from ibeatles.tools.tof_bin.auto_event_handler import AutoEventHandler
from ibeatles.tools.tof_bin.event_handler import EventHandler
from ibeatles.tools.tof_bin.event_handler import EventHandler as TofBinEventHandler
from ibeatles.tools.tof_bin.initialization import Initialization
from ibeatles.tools.tof_bin.manual_event_handler import ManualEventHandler
from ibeatles.tools.tof_bin.manual_right_click import ManualRightClick
from ibeatles.tools.tof_bin.preview_full_bin_axis import PreviewFullBinAxis
from ibeatles.tools.tof_bin.statistics import Statistics
from ibeatles.tools.tof_bin.tof_bin_export_launcher import TofBinExportLauncher
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.tools.utilities.time_spectra import TimeSpectraLauncher

warnings.filterwarnings("ignore")
TIME_SPECTRA_NAME_FORMAT = "*_Spectra.txt"


class TofBinningLauncher:
    def __init__(self, parent=None):
        self.parent = parent

        if self.parent.binning_ui is None:
            tof_combining_binning_window = TofBinning(parent=parent)
            tof_combining_binning_window.setWindowFlags(
                tof_combining_binning_window.windowFlags() | Qt.WindowStaysOnTopHint
            )
            tof_combining_binning_window.show()
            self.parent.tof_combining_binning_ui = tof_combining_binning_window

        else:
            self.parent.binning_ui.setFocus()
            self.parent.binning_ui.activateWindow()


class TofBinning(QMainWindow):
    session = session

    data_array = None  # array of images
    data_array_error = None  # array of images uncertainties

    list_tif_files = None
    images_array = None
    integrated_image = None
    roi_item = None

    integrated_view = None  # pg integrated image (for ROI selection)
    bin_profile_view = None  # pg profile

    image_size = {"width": None, "height": None}

    # time spectra dict
    time_spectra = {
        TimeSpectraKeys.file_name: None,
        TimeSpectraKeys.tof_array: None,
        TimeSpectraKeys.lambda_array: None,
        TimeSpectraKeys.file_index_array: None,
    }

    bin_roi = {"x0": 0, "y0": 0, "width": 100, "height": 100}

    profile_signal = None

    # dictionary of all the bins pg item
    # {0: pg.regionitem1,
    #  2: pg.regionitem2,
    #  ...
    # }
    dict_of_bins_item = None

    linear_bins_selected = None
    log_bins_selected = None

    # each will be a dictionaries of ranges
    # ex: TimeSpectraKeys.tof_array = {0: [1],
    #                                  1: [2,6],
    #                                  3: [7,8,9,10], ...}
    manual_bins = {
        TimeSpectraKeys.tof_array: None,
        TimeSpectraKeys.file_index_array: None,
        TimeSpectraKeys.lambda_array: None,
    }

    # list of manual bins.
    # using a list because any of the bin can be removed by the user
    list_of_manual_bins_item = []

    current_auto_bin_rows_highlighted = []

    # stats currently displayed in the bin stats table
    # {StatisticsName.mean: {Statistics.full: [],
    #                        Statistics.roi: [],
    #                        },
    # StatisticsName.median: ....
    #  }
    current_stats = {BinMode.auto: None, BinMode.manual: None}

    shutter_counts_file = None
    statistics_dict = None  # master statistics dict (preview of all bin stats)

    def __init__(self, parent=None):
        """
        Initialization
        Parameters
        ----------
        """
        super(TofBinning, self).__init__(parent)
        self.ui = load_ui("ui_tof_binning.ui", baseinstance=self)
        self.top_parent = parent

        o_init = Initialization(parent=self, top_parent=parent)
        o_init.all()
        o_init.setup()

        o_event = TofBinEventHandler(parent=self, top_parent=parent)
        o_event.check_widgets()

        self.setWindowTitle("TOF bin")

    # event
    def select_folder_clicked(self):
        self.o_event = TofBinEventHandler(parent=self, top_parent=self.top_parent)
        self.o_event.select_input_folder()
        if self.list_tif_files:
            self.o_event.load_data()

            is_time_spectra_found = self.o_event.load_time_spectra_file()
            if not is_time_spectra_found:
                self.o_event.browse_for_time_spectra_file()

            self.o_event.load_shutter_counts_file()

            self.o_event.display_integrated_image()
            self.o_event.display_profile()
            self.bin_auto_manual_tab_changed()
            self.o_event.check_widgets()

    def bin_roi_changed(self):
        o_event = TofBinEventHandler(parent=self)
        o_event.display_profile()

    def bin_xaxis_changed(self):
        o_event = TofBinEventHandler(parent=self)
        o_event.display_profile()
        o_event.bin_xaxis_changed()
        o_event.check_widgets()

    # - auto mode
    def bin_auto_log_linear_radioButton_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.bin_auto_radioButton_clicked()
        self.update_statistics()

    def time_spectra_preview_clicked(self):
        TimeSpectraLauncher(parent=self)

    def combine_mean_median_changed(self):
        self.update_statistics()

    def roi_selection_profile_tab_changed(self, new_tab_index=-1):
        if new_tab_index == 1:
            self.bin_xaxis_changed()

    def bin_auto_manual_tab_changed(self, new_tab_index=-1):
        if self.images_array:
            o_event = TofBinEventHandler(parent=self)
            o_event.bin_auto_manual_tab_changed(new_tab_index)
            self.update_statistics()
            o_event_global = EventHandler(parent=self)
            o_event_global.check_widgets()
        logging.info("bin_auto_manual_tab_changed completed")

    def bin_auto_log_file_index_changed(self):
        if self.images_array:
            o_event = AutoEventHandler(parent=self)
            o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.file_index_array)
            self.update_statistics()

    def bin_auto_log_tof_changed(self):
        if self.images_array:
            o_event = AutoEventHandler(parent=self)
            o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.tof_array)
            self.update_statistics()

    def bin_auto_log_lambda_changed(self):
        if self.images_array:
            o_event = AutoEventHandler(parent=self)
            o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.lambda_array)
            self.update_statistics()

    def bin_auto_linear_file_index_changed(self):
        if self.images_array:
            o_event = AutoEventHandler(parent=self)
            o_event.clear_selection(auto_mode=BinAutoMode.linear)
            o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.file_index_array)
            self.update_statistics()

    def bin_auto_linear_tof_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.tof_array)
        self.update_statistics()

    def bin_auto_linear_lambda_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.lambda_array)
        self.update_statistics()

    def auto_log_radioButton_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.log)
        o_event.auto_log_radioButton_changed()
        self.update_statistics()

    def auto_linear_radioButton_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.auto_linear_radioButton_changed()
        self.update_statistics()

    def auto_table_use_checkbox_changed(self, state, row):
        o_event = AutoEventHandler(parent=self)
        state = True if state == 2 else False
        o_event.use_auto_bin_state_changed(row=row, state=state)
        self.bin_auto_table_selection_changed()
        self.update_statistics()
        o_event_global = EventHandler(parent=self)
        o_event_global.check_widgets()

    def bin_auto_hide_empty_bins(self):
        o_event = AutoEventHandler(parent=self)
        o_event.update_auto_table()

    def bin_auto_visualize_axis_generated_button_clicked(self):
        o_preview = PreviewFullBinAxis(parent=self)
        o_preview.show()

    def bin_auto_table_right_clicked(self, position):
        o_event = AutoEventHandler(parent=self)
        o_event.auto_table_right_click(position=position)

    def bin_auto_table_selection_changed(self):
        o_event = AutoEventHandler(parent=self)
        o_event.auto_table_selection_changed()

    def mouse_moved_in_combine_image_preview(self):
        """Mouse moved in the combine pyqtgraph image preview (top right)"""
        pass

    # - manual mode
    def bin_manual_add_bin_clicked(self):
        o_event = ManualEventHandler(parent=self)
        o_event.add_bin()
        self.update_statistics()
        o_event_global = EventHandler(parent=self)
        o_event_global.check_widgets()

    def bin_manual_populate_table_with_auto_mode_bins_clicked(self):
        o_event = ManualEventHandler(parent=self)
        o_event.clear_all_items()
        o_event.populate_table_with_auto_mode()
        self.update_statistics()
        o_event_global = EventHandler(parent=self)
        o_event_global.check_widgets()

    def bin_manual_region_changed(self, item_id):
        o_event = ManualEventHandler(parent=self)
        o_event.bin_manually_moved(item_id=item_id)
        self.update_statistics()
        o_event_global = EventHandler(parent=self)
        o_event_global.check_widgets()

    def bin_manual_region_changing(self, item_id):
        o_event = ManualEventHandler(parent=self)
        o_event.bin_manually_moving(item_id=item_id)

    def bin_manual_table_right_click(self, position):
        o_event = ManualRightClick(parent=self)
        o_event.manual_table_right_click()
        o_event_global = EventHandler(parent=self)
        o_event_global.check_widgets()

    # - statistics
    def update_statistics(self):
        o_stat = Statistics(parent=self)
        o_stat.update()
        o_stat.plot_statistics()

    def bin_statistics_comboBox_changed(self):
        o_stat = Statistics(parent=self)
        o_stat.plot_statistics()

    # export images
    def export_bin_images_clicked(self):
        o_export = TofBinExportLauncher(parent=self, top_parent=self.top_parent)
        o_export.show()

    def closeEvent(self, event):
        logging.info(" #### Leaving combine/binning ####")
        self.close()



#!/usr/bin/env python
"""
FitRegions class for fitting the high lambda, low lambda and bragg peak regions.
"""

import copy

import numpy as np
from lmfit import Model, Parameter
from loguru import logger
from qtpy import QtGui

import ibeatles.utilities.error as fitting_error
from ibeatles.fitting import KropffTabSelected
from ibeatles.fitting.kropff import ERROR_TOLERANCE
from ibeatles.fitting.kropff.checking_fitting_conditions import (
    CheckingFittingConditions,
)
from ibeatles.fitting.kropff.fitting_functions import (
    kropff_bragg_peak_tof,
    kropff_high_lambda,
    kropff_low_lambda,
)
from ibeatles.fitting.kropff.get import Get
from ibeatles.utilities.array_utilities import find_nearest_index
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)


class FitRegions:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent
        self.o_get = Get(parent=parent)
        self.table_dictionary = self.grand_parent.kropff_table_dictionary
        self.progress_bar_ui = self.parent.eventProgress

    def all_regions(self):
        type_error = ""

        # o_event = KropffBraggPeakThresholdCalculator(parent=self.parent,
        #                                              grand_parent=self.grand_parent)
        # o_event.save_all_profiles()

        # o_display = Display(parent=self.parent,
        #                     grand_parent=self.grand_parent)
        # o_display.display_bragg_peak_threshold()

        try:
            self.high_lambda()
            self.low_lambda()
            self.bragg_peak()
        except fitting_error.HighLambdaFittingError as err:
            type_error = err
        except fitting_error.LowLambdaFittingError as err:
            type_error = err
        except fitting_error.BraggPeakFittingError as err:
            type_error = err

        if type_error:
            show_status_message(
                parent=self.parent,
                message=f"Error fitting {type_error}",
                status=StatusMessageStatus.error,
                duration_s=5,
            )

    @staticmethod
    def error_outside_of_tolerance(list_error):
        for _error in list_error:
            if not _error:
                return True

            if _error > ERROR_TOLERANCE:
                return True
        return False

    def high_lambda(self):
        gmodel = Model(kropff_high_lambda, missing="drop", independent_vars=["lda"])

        a0 = self.o_get.a0()
        b0 = self.o_get.b0()

        table_dictionary = self.table_dictionary
        # common_xaxis = None
        nearest_index = -1

        self.progress_bar_ui.setMaximum(len(table_dictionary.keys()))
        self.progress_bar_ui.setVisible(True)
        QtGui.QGuiApplication.processEvents()

        xaxis_common = table_dictionary["0"]["xaxis"]

        for _index, _key in enumerate(table_dictionary.keys()):
            self.progress_bar_ui.setValue(_index + 1)
            QtGui.QGuiApplication.processEvents()

            # if row is locked, continue
            if table_dictionary[_key]["lock"]:
                continue

            right = table_dictionary[_key]["bragg peak threshold"]["right"]

            nearest_index = find_nearest_index(xaxis_common, right)
            xaxis = xaxis_common[nearest_index:-1]
            if len(xaxis) == 0:
                table_dictionary[_key]["rejected"] = True
                continue

            yaxis = table_dictionary[_key]["yaxis"]
            yaxis = yaxis[nearest_index:-1]
            yaxis = -np.log(yaxis)

            try:
                _result = gmodel.fit(yaxis, lda=xaxis, a0=a0, b0=b0)
            except ValueError:
                table_dictionary[_key]["rejected"] = True
                continue
            except TypeError:
                table_dictionary[_key]["rejected"] = True
                continue

            a0_value = _result.params["a0"].value
            a0_error = _result.params["a0"].stderr
            b0_value = _result.params["b0"].value
            b0_error = _result.params["b0"].stderr

            if FitRegions.error_outside_of_tolerance([a0_error, b0_error]):
                table_dictionary[_key]["rejected"] = True
                continue

            yaxis_fitted = kropff_high_lambda(xaxis, a0_value, b0_value)

            table_dictionary[_key]["a0"] = {"val": a0_value, "err": a0_error}
            table_dictionary[_key]["b0"] = {"val": b0_value, "err": b0_error}
            table_dictionary[_key]["fitted"][KropffTabSelected.high_tof] = {
                "xaxis": xaxis,
                "yaxis": yaxis_fitted,
            }

        self.progress_bar_ui.setVisible(False)
        QtGui.QGuiApplication.processEvents()

    def low_lambda(self):
        gmodel = Model(kropff_low_lambda, missing="drop", independent_vars=["lda"])

        ahkl = self.o_get.ahkl()
        bhkl = self.o_get.bhkl()

        table_dictionary = self.table_dictionary

        self.progress_bar_ui.setMaximum(len(table_dictionary.keys()))
        self.progress_bar_ui.setVisible(True)
        QtGui.QGuiApplication.processEvents()

        common_xaxis = None
        nearest_index = -1
        for _index, _key in enumerate(table_dictionary.keys()):
            self.progress_bar_ui.setValue(_index + 1)
            QtGui.QGuiApplication.processEvents()

            # if row is locked, continue
            if table_dictionary[_key]["lock"]:
                continue

            if table_dictionary[_key]["rejected"]:
                continue

            table_entry = table_dictionary[_key]

            if nearest_index == -1:
                xaxis = table_entry["xaxis"]

                left = table_entry["bragg peak threshold"]["left"]

                nearest_index = find_nearest_index(xaxis, left)
                xaxis = xaxis[: nearest_index + 1]
                common_xaxis = xaxis

            a0 = table_entry["a0"]["val"]
            b0 = table_entry["b0"]["val"]

            yaxis = table_entry["yaxis"]
            yaxis = yaxis[: nearest_index + 1]
            yaxis = -np.log(yaxis)

            try:
                _result = gmodel.fit(
                    yaxis,
                    lda=common_xaxis,
                    a0=Parameter("a0", value=a0, vary=False),
                    b0=Parameter("b0", value=b0, vary=False),
                    ahkl=ahkl,
                    bhkl=bhkl,
                )
            except ValueError:
                table_dictionary[_key]["rejected"] = True
                continue

            ahkl_value = _result.params["ahkl"].value
            ahkl_error = _result.params["ahkl"].stderr
            bhkl_value = _result.params["bhkl"].value
            bhkl_error = _result.params["bhkl"].stderr
            yaxis_fitted = kropff_low_lambda(common_xaxis, a0, b0, ahkl_value, bhkl_value)

            table_dictionary[_key]["ahkl"] = {"val": ahkl_value, "err": ahkl_error}
            table_dictionary[_key]["bhkl"] = {"val": bhkl_value, "err": bhkl_error}
            table_dictionary[_key]["fitted"][KropffTabSelected.low_tof] = {
                "xaxis": common_xaxis,
                "yaxis": yaxis_fitted,
            }

        self.progress_bar_ui.setVisible(False)
        self.grand_parent.kropff_table_dictionary = self.table_dictionary
        QtGui.QGuiApplication.processEvents()

    def bragg_peak(self):
        self.bragg_peak_range()

        # if self.parent.kropff_lambda_settings['state'] == 'fix':
        #     self.bragg_peak_fix_lambda()
        # else:
        #     self.bragg_peak_range_lambda()

    def bragg_peak_range(self):
        """will fit using either fix or range values of lambda_hkl, tau and sigma"""
        logger.info("Fitting bragg peak with:")

        o_get = Get(parent=self.parent)
        list_lambda_hkl = o_get.list_lambda_hkl_initial_guess()
        list_tau = o_get.list_tau_initial_guess()
        list_sigma = o_get.list_sigma_initial_guess()
        logger.info(f"-> {list_lambda_hkl =}")
        logger.info(f"-> {list_tau =}")
        logger.info(f"-> {list_sigma =}")

        # define fitting model
        gmodel = Model(kropff_bragg_peak_tof, nan_policy="propagate", independent_vars=["lda"])

        table_dictionary = self.table_dictionary
        fit_conditions = self.parent.kropff_bragg_peak_good_fit_conditions
        o_checking = CheckingFittingConditions(fit_conditions=fit_conditions)

        self.parent.eventProgress.setMaximum(len(table_dictionary.keys()))
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)
        QtGui.QGuiApplication.processEvents()

        for _index, _key in enumerate(table_dictionary.keys()):
            # if row is locked, continue
            if table_dictionary[_key]["lock"]:
                self.parent.eventProgress.setValue(_index)
                QtGui.QGuiApplication.processEvents()
                continue

            if table_dictionary[_key]["rejected"]:
                self.parent.eventProgress.setValue(_index)
                QtGui.QGuiApplication.processEvents()
                continue

            table_entry = table_dictionary[_key]

            xaxis = copy.deepcopy(table_entry["xaxis"])

            a0 = table_entry["a0"]["val"]
            b0 = table_entry["b0"]["val"]
            ahkl = table_entry["ahkl"]["val"]
            bhkl = table_entry["bhkl"]["val"]

            yaxis = copy.deepcopy(table_entry["yaxis"])
            yaxis = -np.log(yaxis)

            is_fitting_ok = False
            for _lambda_hkl in list_lambda_hkl:
                for _sigma in list_sigma:
                    for _tau in list_tau:
                        _result = gmodel.fit(
                            yaxis,
                            lda=xaxis,
                            a0=Parameter("a0", value=a0, vary=False),
                            b0=Parameter("b0", value=b0, vary=False),
                            ahkl=Parameter("ahkl", value=ahkl, vary=False),
                            bhkl=Parameter("bhkl", value=bhkl, vary=False),
                            ldahkl=_lambda_hkl,
                            sigma=_sigma,
                            tau=_tau,
                        )

                        ldahkl_error = _result.params["ldahkl"].stderr
                        sigma_error = _result.params["sigma"].stderr
                        tau_error = _result.params["tau"].stderr
                        ldahkl_value = _result.params["ldahkl"].value
                        sigma_value = _result.params["sigma"].value
                        tau_value = _result.params["tau"].value
                        yaxis_fitted = kropff_bragg_peak_tof(
                            xaxis,
                            a0,
                            b0,
                            ahkl,
                            bhkl,
                            ldahkl_value,
                            sigma_value,
                            tau_value,
                        )

                        is_fitting_ok = o_checking.is_fitting_ok(
                            l_hkl_error=ldahkl_error,
                            t_error=tau_error,
                            sigma_error=sigma_error,
                        )
                        self.parent.eventProgress.setValue(_index)
                        QtGui.QGuiApplication.processEvents()

                        table_dictionary[_key]["lambda_hkl"] = {
                            "val": ldahkl_value,
                            "err": ldahkl_error,
                        }
                        table_dictionary[_key]["tau"] = {
                            "val": tau_value,
                            "err": tau_error,
                        }
                        table_dictionary[_key]["sigma"] = {
                            "val": sigma_value,
                            "err": sigma_error,
                        }
                        table_dictionary[_key]["fitted"][KropffTabSelected.bragg_peak] = {
                            "xaxis": xaxis,
                            "yaxis": yaxis_fitted,
                        }

                        if is_fitting_ok:
                            break

                    if is_fitting_ok:
                        break

                if is_fitting_ok:
                    table_dictionary[_key]["lock"] = True
                    break

        self.parent.eventProgress.setVisible(False)

    def bragg_peak_range_lambda(self):
        """we need to try to fit until the fit worked or we exhausted the full range defined"""
        logger.info("Fitting bragg peak with a range of lambda_hkl:")
        gmodel = Model(kropff_bragg_peak_tof, nan_policy="propagate", independent_vars=["lda"])

        # lambda_hkl = self.o_get.lambda_hkl()
        tau = self.o_get.tau()
        sigma = self.o_get.sigma()

        table_dictionary = self.table_dictionary
        fit_conditions = self.parent.kropff_bragg_peak_good_fit_conditions
        o_checking = CheckingFittingConditions(fit_conditions=fit_conditions)

        lambda_hkl_range = np.arange(
            self.parent.kropff_lambda_settings["range"][0],
            self.parent.kropff_lambda_settings["range"][1],
            self.parent.kropff_lambda_settings["range"][2],
        )
        logger.info(f"-> lambda_hkl_range: {lambda_hkl_range}")

        self.parent.eventProgress.setMaximum(len(table_dictionary.keys()))
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)
        QtGui.QGuiApplication.processEvents()

        for _index, _key in enumerate(table_dictionary.keys()):
            # if row is locked, continue
            if table_dictionary[_key]["lock"]:
                self.parent.eventProgress.setValue(_index)
                QtGui.QGuiApplication.processEvents()
                continue

            if table_dictionary[_key]["rejected"]:
                self.parent.eventProgress.setValue(_index)
                QtGui.QGuiApplication.processEvents()
                continue

            table_entry = table_dictionary[_key]

            xaxis = copy.deepcopy(table_entry["xaxis"])

            a0 = table_entry["a0"]["val"]
            b0 = table_entry["b0"]["val"]
            ahkl = table_entry["ahkl"]["val"]
            bhkl = table_entry["bhkl"]["val"]

            yaxis = copy.deepcopy(table_entry["yaxis"])
            yaxis = -np.log(yaxis)

            for _lambda_hkl in lambda_hkl_range:
                _result = gmodel.fit(
                    yaxis,
                    lda=xaxis,
                    a0=Parameter("a0", value=a0, vary=False),
                    b0=Parameter("b0", value=b0, vary=False),
                    ahkl=Parameter("ahkl", value=ahkl, vary=False),
                    bhkl=Parameter("bhkl", value=bhkl, vary=False),
                    ldahkl=_lambda_hkl,
                    sigma=sigma,
                    tau=tau,
                )

                ldahkl_error = _result.params["ldahkl"].stderr
                sigma_error = _result.params["sigma"].stderr
                tau_error = _result.params["tau"].stderr
                ldahkl_value = _result.params["ldahkl"].value
                sigma_value = _result.params["sigma"].value
                tau_value = _result.params["tau"].value
                yaxis_fitted = kropff_bragg_peak_tof(xaxis, a0, b0, ahkl, bhkl, ldahkl_value, sigma_value, tau_value)

                if o_checking.is_fitting_ok(l_hkl_error=ldahkl_error, t_error=tau_error, sigma_error=sigma_error):
                    break

            self.parent.eventProgress.setValue(_index)
            QtGui.QGuiApplication.processEvents()

            table_dictionary[_key]["lambda_hkl"] = {
                "val": ldahkl_value,
                "err": ldahkl_error,
            }
            table_dictionary[_key]["tau"] = {"val": tau_value, "err": tau_error}
            table_dictionary[_key]["sigma"] = {"val": sigma_value, "err": sigma_error}
            table_dictionary[_key]["fitted"][KropffTabSelected.bragg_peak] = {
                "xaxis": xaxis,
                "yaxis": yaxis_fitted,
            }
        self.parent.eventProgress.setVisible(False)

    def bragg_peak_fix_lambda(self):
        logger.info("Fitting bragg peak with a fixed initial lambda_hkl:")

        gmodel = Model(kropff_bragg_peak_tof, nan_policy="propagate", independent_vars=["lda"])

        lambda_hkl = self.o_get.lambda_hkl()
        logger.info(f"-> lambda_hkl: {lambda_hkl}")
        tau = self.o_get.tau()
        sigma = self.o_get.sigma()

        table_dictionary = self.table_dictionary

        self.progress_bar_ui.setMaximum(len(table_dictionary.keys()))
        self.progress_bar_ui.setVisible(True)
        QtGui.QGuiApplication.processEvents()

        list_key_locked = []
        for _index, _key in enumerate(table_dictionary.keys()):
            self.progress_bar_ui.setValue(_index + 1)
            QtGui.QGuiApplication.processEvents()

            # if row is locked, continue
            if table_dictionary[_key]["lock"]:
                list_key_locked.append(int(_key) + 1)
                continue

            if table_dictionary[_key]["rejected"]:
                continue

            table_entry = table_dictionary[_key]

            xaxis = copy.deepcopy(table_entry["xaxis"])

            a0 = table_entry["a0"]["val"]
            b0 = table_entry["b0"]["val"]
            ahkl = table_entry["ahkl"]["val"]
            bhkl = table_entry["bhkl"]["val"]

            yaxis = copy.deepcopy(table_entry["yaxis"])
            yaxis = -np.log(yaxis)

            _result = gmodel.fit(
                yaxis,
                lda=xaxis,
                a0=Parameter("a0", value=a0, vary=False),
                b0=Parameter("b0", value=b0, vary=False),
                ahkl=Parameter("ahkl", value=ahkl, vary=False),
                bhkl=Parameter("bhkl", value=bhkl, vary=False),
                ldahkl=lambda_hkl,
                sigma=sigma,
                tau=tau,
            )

            if _key in ["2", "3"]:
                logger.info(f"_key: {_key} -> _result.fit_report: {_result.fit_report()}")

            ldahkl_value = _result.params["ldahkl"].value
            ldahkl_error = _result.params["ldahkl"].stderr
            sigma_value = _result.params["sigma"].value
            sigma_error = _result.params["sigma"].stderr
            tau_value = _result.params["tau"].value
            tau_error = _result.params["tau"].stderr
            yaxis_fitted = kropff_bragg_peak_tof(xaxis, a0, b0, ahkl, bhkl, ldahkl_value, sigma_value, tau_value)

            table_dictionary[_key]["lambda_hkl"] = {
                "val": ldahkl_value,
                "err": ldahkl_error,
            }
            table_dictionary[_key]["tau"] = {"val": tau_value, "err": tau_error}
            table_dictionary[_key]["sigma"] = {"val": sigma_value, "err": sigma_error}
            table_dictionary[_key]["fitted"][KropffTabSelected.bragg_peak] = {
                "xaxis": xaxis,
                "yaxis": yaxis_fitted,
            }

        logger.info(f"-> list of locked keys: {list_key_locked}")

        self.progress_bar_ui.setVisible(False)
        QtGui.QGuiApplication.processEvents()



from copy import deepcopy


class RoiHandler:
    def __init__(self, parent=None, data_type="sample"):
        self.parent = parent
        self.data_type = data_type

    def get_roi_index_that_changed(self):
        old_list_roi = self.parent.old_list_roi[self.data_type]
        new_list_roi = self.parent.list_roi[self.data_type]

        if not (len(old_list_roi) == len(new_list_roi)):
            self.parent.old_list_roi[self.data_type] = deepcopy(new_list_roi)

            return -1

        roi_index = -1

        for _index, _roi in enumerate(new_list_roi):
            _previous_roi = old_list_roi[_index]

            if roi_index == -1:
                if self.are_array_not_equal(_roi, _previous_roi):
                    roi_index = _index
                    old_list_roi[_index] = [_roi[:]]
                    break

        self.parent.old_list_roi[self.data_type] = old_list_roi

        return roi_index

    def are_array_not_equal(self, array1, array2):
        for _index, _value in enumerate(array1):
            if not (_value == array2[_index]):
                return False

        return True



from ibeatles import RegionType

roi_label_color = {
    RegionType.sample: "#0000ff",  # blue
    RegionType.background: "#ff0000",
}  # red


class KernelType:
    box = "Box"
    gaussian = "Gaussian"



#!/usr/bin/env python
"""Detector-specific modules for neutron imaging data processing.

This package contains modules for TPX1 MCP detector uncertainty calculations.
"""

from ibeatles.core.detector.uncertainty import (
    compute_counts_with_uncertainty,
    load_shutter_counts,
)

__all__ = [
    "compute_counts_with_uncertainty",
    "load_shutter_counts",
]



#!/usr/bin/env python
"""
SaveFittingTab class
"""

from loguru import logger

from ibeatles import DataType
from ibeatles.fitting import FittingKeys, FittingTabSelected
from ibeatles.fitting.kropff import SessionSubKeys as KropffSessionSubKeys
from ibeatles.fitting.march_dollase import (
    SessionSubKeys as MarchDollaseSessionSubKeys,
)
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.session.save_tab import SaveTab
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class SaveFittingTab(SaveTab):
    def fitting(self):
        if not self.parent.session_dict[DataType.fitting][SessionSubKeys.ui_accessed]:
            return

        self.general_infos()
        self.march_dollase()
        self.kropff()

    def general_infos(self):
        logger.info("Recording general fitting parameters")

        if self.parent.fitting_image_view:
            o_pyqt = PyqtgraphUtilities(
                parent=self.parent,
                image_view=self.parent.fitting_image_view,
                data_type=DataType.fitting,
            )
            state = o_pyqt.get_state()
            o_pyqt.save_histogram_level(data_type_of_data=DataType.normalized)
            histogram = self.parent.image_view_settings[DataType.fitting]["histogram"]
        else:
            state = None
            histogram = None

        fitting_bragg_edge_linear_selection = self.parent.fitting_bragg_edge_linear_selection
        if fitting_bragg_edge_linear_selection:
            min_lambda_index = int(fitting_bragg_edge_linear_selection[0])
            max_lambda_index = int(fitting_bragg_edge_linear_selection[1])
            self.session_dict[DataType.fitting]["lambda range index"] = [
                min_lambda_index,
                max_lambda_index,
            ]

        logger.info(f" state: {state}")
        logger.info(f" histogram: {histogram}")

        self.session_dict[DataType.fitting][FittingKeys.x_axis] = [
            float(x) for x in self.parent.normalized_lambda_bragg_edge_x_axis
        ]
        self.session_dict[DataType.fitting][FittingKeys.transparency] = self.parent.fitting_transparency_slider_value
        self.session_dict[DataType.fitting][FittingKeys.image_view_state] = state
        self.session_dict[DataType.fitting][FittingKeys.image_view_histogram] = histogram
        self.session_dict[DataType.fitting][FittingKeys.ui_accessed] = self.parent.session_dict[DataType.fitting][
            SessionSubKeys.ui_accessed
        ]
        self.session_dict[DataType.fitting][FittingKeys.ui] = self.parent.session_dict[DataType.fitting]["ui"]

    def march_dollase(self):
        logger.info("Recording March-Dollase fitting parameters")

        if self.parent.fitting_ui:
            self.parent.fitting_ui.save_all_parameters()

        table_dictionary = self.parent.march_table_dictionary

        formatted_table_dictionary = {}

        for _row in table_dictionary.keys():
            _entry = table_dictionary[_row]

            active_flag = _entry["active"]
            lock_flag = _entry["lock"]
            fitting_confidence = _entry["fitting_confidence"]
            d_spacing = _entry["d_spacing"]
            sigma = _entry["sigma"]
            alpha = _entry["alpha"]
            a1 = _entry["a1"]
            a2 = _entry["a2"]
            a5 = _entry["a5"]
            a6 = _entry["a6"]

            formatted_table_dictionary[_row] = {
                "active": active_flag,
                "lock": lock_flag,
                "fitting_confidence": fitting_confidence,
                "d_spacing": d_spacing,
                "sigma": sigma,
                "alpha": alpha,
                "a1": a1,
                "a2": a2,
                "a5": a5,
                "a6": a6,
            }
        self.session_dict[DataType.fitting][FittingTabSelected.march_dollase][
            MarchDollaseSessionSubKeys.table_dictionary
        ] = formatted_table_dictionary
        self.session_dict[DataType.fitting][FittingTabSelected.march_dollase][
            MarchDollaseSessionSubKeys.plot_active_row_flag
        ] = self.parent.display_active_row_flag

        x_axis = self.session_dict[DataType.fitting][SessionSubKeys.x_axis]
        if x_axis:
            logger.info(f" len(x_axis): {len(x_axis)}")
        else:
            logger.info(" x_axis is empty!")

        logger.info(f" lambda range index: {self.session_dict[SessionKeys.fitting][SessionSubKeys.lambda_range_index]}")

    def kropff(self):
        logger.info("Recording Kropff fitting parameters")
        table_dictionary = self.parent.kropff_table_dictionary

        formatted_table_dictionary = {}

        for _row in table_dictionary.keys():
            _entry = table_dictionary[_row]

            a0 = _entry["a0"]
            b0 = _entry["b0"]
            ahkl = _entry["ahkl"]
            bhkl = _entry["bhkl"]
            lambda_hkl = _entry["lambda_hkl"]
            tau = _entry["tau"]
            sigma = _entry["sigma"]
            bragg_peak_threshold = _entry["bragg peak threshold"]
            lock = _entry["lock"]
            rejected = _entry["rejected"]
            row_index = _entry[FittingKeys.row_index]
            column_index = _entry[FittingKeys.column_index]

            formatted_table_dictionary[_row] = {
                "a0": a0,
                "b0": b0,
                "ahkl": ahkl,
                "bhkl": bhkl,
                "lambda_hkl": lambda_hkl,
                "tau": tau,
                "sigma": sigma,
                "bragg_peak_threshold": bragg_peak_threshold,
                "lock": lock,
                "rejected": rejected,
                FittingKeys.row_index: row_index,
                FittingKeys.column_index: column_index,
            }

        self.import_from_parent_session_dict(
            key=KropffSessionSubKeys.table_dictionary, source=formatted_table_dictionary
        )
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.automatic_bragg_peak_threshold_finder)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.automatic_bragg_peak_threshold_algorithm)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.high_tof)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.low_tof)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.bragg_peak)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.kropff_bragg_peak_good_fit_conditions)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.kropff_lambda_settings)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.bragg_peak_row_rejections_conditions)
        self.import_from_parent_session_dict(key=KropffSessionSubKeys.automatic_fitting_threshold_width)

    def import_from_parent_session_dict(self, key: SessionSubKeys = None, source: dict = None):
        """
        this method will move the key values specified from the self.parent.session_dict[fitting][kropff][key]
        if source is not specified, otherwise the source is used as input
        """
        if source is None:
            source = self.parent.session_dict[DataType.fitting][FittingTabSelected.kropff][key]

        self.session_dict[DataType.fitting][FittingTabSelected.kropff][key] = source



#!/usr/bin/env python
import os
import tempfile

# from . import icons_rc  # @UnusedImport
# set the default backend to be compatible with Qt in case someone uses pylab from IPython console
import matplotlib.cm
import matplotlib.colors
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT

# import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.cbook import Stack
from matplotlib.colors import LogNorm, Normalize
from matplotlib.figure import Figure
from qtpy import QtCore, QtGui
from qtpy.QtWidgets import QWidget

try:
    import matplotlib.backends.qt4_editor.figureoptions as figureoptions
except ImportError:
    figureoptions = None

cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
    "default", ["#0000ff", "#00ff00", "#ffff00", "#ff0000", "#bd7efc", "#000000"], N=256
)
matplotlib.cm.register_cmap("default", cmap=cmap)


class NavigationToolbar(NavigationToolbar2QT):
    """
    A small change to the original navigation toolbar.
    """

    _auto_toggle = False
    logtog = QtCore.pyqtSignal(str)
    homeClicked = QtCore.pyqtSignal()
    exportClicked = QtCore.pyqtSignal()

    isPanActivated = False
    isZoomActivated = False

    home_settings = None

    def __init__(self, canvas, parent, coordinates=False):
        NavigationToolbar2QT.__init__(self, canvas, parent, coordinates)
        self.setIconSize(QtCore.QSize(20, 20))

    def _init_toolbar(self):
        if not hasattr(self, "_actions"):
            self._actions = {}
        self.basedir = os.path.join(matplotlib.rcParams["datapath"], "images")

        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/go-home.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Home", self.home)
        a.setToolTip("Reset original view")
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/zoom-previous.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Back", self.back)
        a.setToolTip("Back to previous view")
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/zoom-next.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Forward", self.forward)
        a.setToolTip("Forward to next view")
        self.addSeparator()
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/transform-move.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Pan", self.pan)
        a.setToolTip("Pan axes with left mouse, zoom with right")
        a.setCheckable(True)
        self._actions["pan"] = a
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/zoom-select.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Zoom", self.zoom)
        a.setToolTip("Zoom to rectangle")
        a.setCheckable(True)
        self._actions["zoom"] = a
        self.addSeparator()
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/edit-guides.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Subplots", self.configure_subplots)
        a.setToolTip("Configure plot boundaries")

        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/MPL Toolbar/document-save.png"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        a = self.addAction(icon, "Save", self.save_figure)
        a.setToolTip("Save the figure")

        # icon=QtGui.QIcon()
        # icon.addPixmap(QtGui.QPixmap(":/MPL Toolbar/document-print.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        # a=self.addAction(icon, 'Print', self.print_figure)
        # a.setToolTip('Print the figure with the default printer')

        self.buttons = {}

        # Add the x,y location widget at the right side of the toolbar
        # The stretch factor is 1 which means any resizing of the toolbar
        # will resize this label instead of the buttons.
        self.locLabel = QtGui.QLabel("", self)
        self.locLabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignTop)
        self.locLabel.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Ignored))
        self.labelAction = self.addWidget(self.locLabel)
        if self.coordinates:
            self.labelAction.setVisible(True)
        else:
            self.labelAction.setVisible(False)

        # reference holder for subplots_adjust window
        self.adj_window = None

    def activate_widget(self, widget_name, activateIt):
        if widget_name == "pan":
            if activateIt:
                self.isPanActivated = True
                self.isZoomActivated = False
            else:
                self.isPanActivated = False
        elif widget_name == "zoom":
            if activateIt:
                self.isZoomActivated = True
                self.isPanActivated = False
            else:
                self.isZoomActivated = False

    def home(self, *args):
        NavigationToolbar2QT.home(self, *args)
        self.homeClicked.emit()

    def pan(self, *args):
        NavigationToolbar2QT.pan(self, *args)
        self.activate_widget("pan", not self.isPanActivated)

    def zoom(self, *args):
        NavigationToolbar2QT.zoom(self, *args)
        self.activate_widget("zoom", not self.isZoomActivated)

    if matplotlib.__version__ < "1.2":

        def pan(self, *args):  # noqa: F811
            """Activate the pan/zoom tool. pan with left button, zoom with right"""
            # set the pointer icon and button press funcs to the
            # appropriate callbacks
            if self._auto_toggle:
                return
            if self._active == "ZOOM":
                self._auto_toggle = True
                self._actions["zoom"].setChecked(False)
                self._auto_toggle = False

            if self._active == "PAN":
                self._active = None
            else:
                self._active = "PAN"
            if self._idPress is not None:
                self._idPress = self.canvas.mpl_disconnect(self._idPress)
                self.mode = ""

            if self._idRelease is not None:
                self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
                self.mode = ""

            if self._active:
                self._idPress = self.canvas.mpl_connect("button_press_event", self.press_pan)
                self._idRelease = self.canvas.mpl_connect("button_release_event", self.release_pan)
                self.mode = "pan/zoom"
                self.canvas.widgetlock(self)
            else:
                self.canvas.widgetlock.release(self)

            for a in self.canvas.figure.get_axes():
                a.set_navigate_mode(self._active)

            self.set_message(self.mode)

        def zoom(self, *args):  # noqa: F811
            """activate zoom to rect mode"""
            if self._auto_toggle:
                return
            if self._active == "PAN":
                self._auto_toggle = True
                self._actions["pan"].setChecked(False)
                self._auto_toggle = False

            if self._active == "ZOOM":
                self._active = None
            else:
                self._active = "ZOOM"

            if self._idPress is not None:
                self._idPress = self.canvas.mpl_disconnect(self._idPress)
                self.mode = ""

            if self._idRelease is not None:
                self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
                self.mode = ""

            if self._active:
                self._idPress = self.canvas.mpl_connect("button_press_event", self.press_zoom)
                self._idRelease = self.canvas.mpl_connect("button_release_event", self.release_zoom)
                self.mode = "zoom rect"
                self.canvas.widgetlock(self)
            else:
                self.canvas.widgetlock.release(self)

            for a in self.canvas.figure.get_axes():
                a.set_navigate_mode(self._active)

            self.set_message(self.mode)

    def export_ascii(self):
        self.exportClicked.emit()

    def print_figure(self):
        """
        Save the plot to a temporary png file and show a preview dialog also used for printing.
        """
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = filetypes.items()
        sorted_filetypes.sort()

        filename = os.path.join(tempfile.gettempdir(), "quicknxs_print.png")
        self.canvas.print_figure(filename, dpi=600)
        imgpix = QtGui.QPixmap(filename)
        os.remove(filename)

        imgobj = QtGui.QLabel()
        imgobj.setPixmap(imgpix)
        imgobj.setMask(imgpix.mask())
        imgobj.setGeometry(0, 0, imgpix.width(), imgpix.height())

        def getPrintData(printer):
            imgobj.render(printer)

        printer = QtGui.QPrinter()
        printer.setPrinterName("mrac4a_printer")
        printer.setPageSize(QtGui.QPrinter.Letter)
        printer.setResolution(600)
        printer.setOrientation(QtGui.QPrinter.Landscape)

        pd = QtGui.QPrintPreviewDialog(printer)
        pd.paintRequested.connect(getPrintData)
        pd.exec_()

    def save_figure(self, *args):
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = filetypes.items()
        sorted_filetypes.sort()
        default_filetype = self.canvas.get_default_filetype()

        start = "image." + default_filetype
        filters = []
        for name, exts in sorted_filetypes:
            exts_list = " ".join(["*.%s" % ext for ext in exts])
            filter_ = "%s (%s)" % (name, exts_list)
            if default_filetype in exts:
                filters.insert(0, filter_)
            else:
                filters.append(filter_)
        filters = ";;".join(filters)

        fname = QtGui.QFileDialog.getSaveFileName(self, "Choose a filename to save to", start, filters)
        if fname:
            try:
                self.canvas.print_figure(str(fname))
            except Exception as e:
                QtGui.QMessageBox.critical(
                    self,
                    "Error saving file",
                    str(e),
                    QtGui.QMessageBox.Ok,
                    QtGui.QMessageBox.NoButton,
                )

    def toggle_log(self, *args):
        ax = self.canvas.ax
        if len(ax.images) == 0 and all([c.__class__.__name__ != "QuadMesh" for c in ax.collections]):
            logstate = ax.get_yscale()
            if logstate == "linear":
                ax.set_yscale("log")
            else:
                ax.set_yscale("linear")
            self.canvas.draw()
            self.logtog.emit(ax.get_yscale())
        else:
            imgs = ax.images + [c for c in ax.collections if c.__class__.__name__ == "QuadMesh"]
            norm = imgs[0].norm
            if norm.__class__ is LogNorm:
                for img in imgs:
                    img.set_norm(Normalize(norm.vmin, norm.vmax))
            else:
                for img in imgs:
                    img.set_norm(LogNorm(norm.vmin, norm.vmax))
        self.canvas.draw()


class MplCanvas(FigureCanvas):
    from matplotlib import cm

    trigger_click = QtCore.pyqtSignal()
    trigger_figure_left = QtCore.pyqtSignal()

    def __init__(
        self,
        parent=None,
        width=3,
        height=3,
        dpi=100,
        sharex=None,
        sharey=None,
        adjust={},
    ):
        self.fig = Figure(figsize=(width, height), dpi=dpi, facecolor="#FFFFFF")
        self.ax = self.fig.add_subplot(111, sharex=sharex, sharey=sharey)
        self.fig.subplots_adjust(left=0.15, bottom=0.1, right=0.95, top=0.95)
        self.xtitle = ""
        self.ytitle = ""
        self.PlotTitle = ""
        self.grid_status = True
        self.xaxis_style = "linear"
        self.yaxis_style = "linear"
        self.format_labels()
        self.ax.hold(True)
        FigureCanvas.__init__(self, self.fig)
        # self.fc = FigureCanvas(self.fig)
        FigureCanvas.setSizePolicy(self, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

        self.fig.canvas.mpl_connect("button_press_event", self.button_pressed)
        self.fig.canvas.mpl_connect("figure_leave_event", self.figure_leave)

    def button_pressed(self, event):
        self.trigger_click.emit()

    def figure_leave(self, event):
        self.trigger_figure_left.emit()

    def format_labels(self):
        self.ax.set_title(self.PlotTitle)

    #    self.ax.title.set_fontsize(10)
    #    self.ax.set_xlabel(self.xtitle, fontsize=9)
    #    self.ax.set_ylabel(self.ytitle, fontsize=9)
    #    labels_x=self.ax.get_xticklabels()
    #    labels_y=self.ax.get_yticklabels()
    #
    #    for xlabel in labels_x:
    #      xlabel.set_fontsize(8)
    #    for ylabel in labels_y:
    #      ylabel.set_fontsize(8)

    def sizeHint(self):
        w, h = self.get_width_height()
        w = max(w, self.height())
        h = max(h, self.width())
        return QtCore.QSize(w, h)

    def minimumSizeHint(self):
        return QtCore.QSize(40, 40)

    def get_default_filetype(self):
        return "png"


class MPLWidgetColorbar(QWidget):
    cplot = None
    cbar = None

    logtogy = QtCore.pyqtSignal(str)
    singleClick = QtCore.pyqtSignal(bool)
    leaveFigure = QtCore.pyqtSignal()

    def __init__(self, parent=None, with_toolbar=True, coordinates=False):
        QWidget.__init__(self, parent)
        self.canvas = MplCanvas()
        self.canvas.ax2 = None
        self.vbox = QtGui.QVBoxLayout()
        self.vbox.setMargin(1)
        self.vbox.addWidget(self.canvas)
        if with_toolbar:
            self.toolbar = NavigationToolbar(self.canvas, self)
            self.toolbar.coordinates = coordinates
            self.vbox.addWidget(self.toolbar)
            self.toolbar.logtog.connect(self.logtoggleylog)
        else:
            self.toolbar = None
        self.setLayout(self.vbox)
        self.canvas.trigger_click.connect(self._singleClick)
        self.canvas.trigger_figure_left.connect(self._leaveFigure)

    def _singleClick(self):
        status = self.toolbar.isPanActivated or self.toolbar.isZoomActivated
        self.singleClick.emit(status)

    def _leaveFigure(self):
        self.leaveFigure.emit()

    def logtoggleylog(self, status):
        self.logtogy.emit(status)

    def leaveEvent(self, event):
        """
        Make sure the cursor is reset to it's default when leaving the widget.
        In some cases the zoom cursor does not reset when leaving the plot.
        """
        if self.toolbar:
            QtGui.QApplication.restoreOverrideCursor()
            self.toolbar._lastCursor = None
        return QWidget.leaveEvent(self, event)

    def set_config(self, config):
        self.canvas.fig.subplots_adjust(**config)

    def get_config(self):
        spp = self.canvas.fig.subplotpars
        config = dict(left=spp.left, right=spp.right, bottom=spp.bottom, top=spp.top)
        return config

    def draw(self):
        """
        Convenience to redraw the graph.
        """
        self.canvas.draw()

    def plot(self, *args, **opts):
        """
        Convenience wrapper for self.canvas.ax.plot
        """
        return self.canvas.ax.plot(*args, **opts)

    def semilogy(self, *args, **opts):
        """
        Convenience wrapper for self.canvas.ax.semilogy
        """
        return self.canvas.ax.semilogy(*args, **opts)

    def errorbar(self, *args, **opts):
        """
        Convenience wrapper for self.canvas.ax.semilogy
        """
        return self.canvas.ax.errorbar(*args, **opts)

    def pcolormesh(self, datax, datay, dataz, log=False, imin=None, imax=None, update=False, **opts):
        """
        Convenience wrapper for self.canvas.ax.plot
        """
        if self.cplot is None or not update:
            if log:
                self.cplot = self.canvas.ax.pcolormesh(datax, datay, dataz, norm=LogNorm(imin, imax), **opts)
            else:
                self.cplot = self.canvas.ax.pcolormesh(datax, datay, dataz, **opts)
        else:
            self.update(datax, datay, dataz)
        return self.cplot

    def imshow(self, data, log=False, imin=None, imax=None, update=True, **opts):
        """
        Convenience wrapper for self.canvas.ax.plot
        """
        if self.cplot is None or not update:
            if log:
                self.cplot = self.canvas.ax.imshow(data, norm=LogNorm(imin, imax), **opts)
            else:
                self.cplot = self.canvas.ax.imshow(data, **opts)
        else:
            self.update(data, **opts)
        return self.cplot

    def set_title(self, new_title):
        return self.canvas.ax.title.set_text(new_title)

    def set_xlabel(self, label):
        return self.canvas.ax.set_xlabel(label)

    def set_ylabel(self, label):
        return self.canvas.ax.set_ylabel(label)

    def set_xscale(self, scale):
        try:
            return self.canvas.ax.set_xscale(scale)
        except ValueError:
            pass

    def set_yscale(self, scale):
        try:
            return self.canvas.ax.set_yscale(scale)
        except ValueError:
            pass

    def clear_fig(self):
        self.cplot = None
        self.cbar = None
        self.canvas.fig.clear()
        self.canvas.ax = self.canvas.fig.add_subplot(111, sharex=None, sharey=None)

    def clear(self):
        self.cplot = None
        self.toolbar._views.clear()
        self.toolbar._positions.clear()
        self.canvas.ax.clear()
        if self.canvas.ax2 is not None:
            self.canvas.ax2.clear()

    def update(self, *data, **opts):
        self.cplot.set_data(*data)
        if "extent" in opts:
            self.cplot.set_extent(opts["extent"])
            oldviews = self.toolbar._views
            if self.toolbar._views:
                # set the new extent as home for the new data
                newviews = Stack()
                newviews.push([tuple(opts["extent"])])
                for item in oldviews[1:]:
                    newviews.push(item)
                self.toolbar._views = newviews
            if not oldviews or oldviews[oldviews._pos] == oldviews[0]:
                self.canvas.ax.set_xlim(opts["extent"][0], opts["extent"][1])
                self.canvas.ax.set_ylim(opts["extent"][2], opts["extent"][3])

    def legend(self, *args, **opts):
        return self.canvas.ax.legend(*args, **opts)

    def adjust(self, **adjustment):
        return self.canvas.fig.subplots_adjust(**adjustment)



class ParametersToDisplay:
    d = "d"
    strain_mapping = "strain_mapping"
    integrated_image = "integrated_image"


INTERPOLATION_METHODS = [
    "none",
    "nearest",
    "bilinear",
    "bicubic",
    "spline16",
    "spline36",
    "hanning",
    "hamming",
    "hermite",
    "kaiser",
    "quadric",
    "catrom",
    "gaussian",
    "bessel",
    "mitchell",
    "sinc",
    "lanczos",
]


DEFAULT_INTERPOLATION_INDEX = 0


CMAPS = ["viridis", "jet"]

DEFAULT_CMAP_INDEX = 0



#!/usr/bin/env python
"""
AdvancedSelectionLauncher class for launching the advanced selection tool.
"""

import numpy as np
from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QMainWindow, QTableWidgetSelectionRange

from ibeatles import load_ui
from ibeatles.fitting.filling_table_handler import FillingTableHandler
from ibeatles.fitting.march_dollase.event_handler import EventHandler


class AdvancedSelectionLauncher(object):
    def __init__(self, grand_parent=None, parent=None):
        self.grand_parent = grand_parent
        self.parent = parent

        if self.grand_parent.advanced_selection_ui is None:
            advanced_window = AdvancedSelectionWindow(grand_parent=grand_parent, parent=parent)
            self.grand_parent.advanced_selection_ui = advanced_window
            advanced_window.show()
        else:
            self.grand_parent.advanced_selection_ui.setFocus()
            self.grand_parent.advanced_selection_ui.activateWindow()


class AdvancedSelectionWindow(QMainWindow):
    def __init__(self, grand_parent=None, parent=None):
        self.grand_parent = grand_parent
        self.parent = parent
        QMainWindow.__init__(self, parent=grand_parent)
        self.ui = load_ui("ui_advancedFittingSelection.ui", baseinstance=self)
        self.setWindowTitle("Graphical Selection Tool")

        self.ui.selection_table.blockSignals(True)
        self.ui.lock_table.blockSignals(True)
        self.init_table()
        self.ui.selection_table.blockSignals(False)
        self.ui.lock_table.blockSignals(False)

    def init_table(self):
        fitting_selection = self.grand_parent.fitting_selection
        nbr_row = fitting_selection["nbr_row"]
        nbr_column = fitting_selection["nbr_column"]

        # selection table
        self.ui.selection_table.setColumnCount(nbr_column)
        self.ui.selection_table.setRowCount(nbr_row)
        self.update_selection_table()

        # lock table
        self.ui.lock_table.setColumnCount(nbr_column)
        self.ui.lock_table.setRowCount(nbr_row)
        self.update_lock_table()

        # set size of cells
        value = int(self.ui.advanced_selection_cell_size_slider.value())
        self.selection_cell_size_changed(value)

    def update_selection_table(self):
        self.update_table(state_field="active", table_ui=self.ui.selection_table)

    def update_lock_table(self):
        self.update_table(state_field="lock", table_ui=self.ui.lock_table)

    def update_table(self, state_field="", table_ui=None):
        table_dictionary = self.grand_parent.march_table_dictionary

        for _index in table_dictionary:
            _entry = table_dictionary[_index]
            state = _entry[state_field]
            row_index = _entry["row_index"]
            column_index = _entry["column_index"]
            _selection = QTableWidgetSelectionRange(row_index, column_index, row_index, column_index)
            table_ui.setRangeSelected(_selection, state)

    def selection_cell_size_changed(self, value):
        nbr_row = self.ui.selection_table.rowCount()
        nbr_column = self.ui.selection_table.columnCount()

        for _row in np.arange(nbr_row):
            self.ui.selection_table.setRowHeight(_row, value)
            self.ui.lock_table.setRowHeight(_row, value)

        for _col in np.arange(nbr_column):
            self.ui.selection_table.setColumnWidth(_col, value)
            self.ui.lock_table.setColumnWidth(_col, value)

    def selection_table_selection_changed(self):
        # update table and then update GUI
        selection = self.ui.selection_table.selectedRanges()
        nbr_row = self.ui.selection_table.rowCount()

        table_dictionary = self.grand_parent.march_table_dictionary

        for _entry in table_dictionary.keys():
            table_dictionary[_entry]["active"] = False

        for _select in selection:
            top_row = _select.topRow()
            left_col = _select.leftColumn()
            bottom_row = _select.bottomRow()
            right_col = _select.rightColumn()
            for _row in np.arange(top_row, bottom_row + 1):
                for _col in np.arange(left_col, right_col + 1):
                    fitting_row = _col * nbr_row + _row
                    _entry = table_dictionary[str(fitting_row)]
                    _entry["active"] = True
                    table_dictionary[str(fitting_row)] = _entry

        self.grand_parent.march_table_dictionary = table_dictionary
        o_filling_table = FillingTableHandler(grand_parent=self.grand_parent, parent=self.parent)

        self.grand_parent.fitting_ui.ui.value_table.blockSignals(True)
        o_filling_table.fill_table()
        self.grand_parent.fitting_ui.ui.value_table.blockSignals(False)

        o_event = EventHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_event.update_image_view_selection()

        self.grand_parent.fitting_ui.update_bragg_edge_plot()

    def lock_table_selection_changed(self):
        # update table and then update GUI
        selection = self.ui.lock_table.selectedRanges()
        nbr_row = self.ui.lock_table.rowCount()

        table_dictionary = self.grand_parent.march_table_dictionary

        for _entry in table_dictionary.keys():
            table_dictionary[_entry]["lock"] = False

        for _select in selection:
            top_row = _select.topRow()
            left_col = _select.leftColumn()
            bottom_row = _select.bottomRow()
            right_col = _select.rightColumn()
            for _row in np.arange(top_row, bottom_row + 1):
                for _col in np.arange(left_col, right_col + 1):
                    fitting_row = _col * nbr_row + _row
                    _entry = table_dictionary[str(fitting_row)]
                    _entry["lock"] = True
                    table_dictionary[str(fitting_row)] = _entry

        self.grand_parent.march_table_dictionary = table_dictionary
        o_filling_table = FillingTableHandler(grand_parent=self.grand_parent, parent=self.parent)

        self.grand_parent.fitting_ui.ui.value_table.blockSignals(True)
        o_filling_table.fill_table()
        self.grand_parent.fitting_ui.ui.value_table.blockSignals(False)

        o_event = EventHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_event.update_image_view_lock()

        self.grand_parent.fitting_ui.update_bragg_edge_plot()

        if self.grand_parent.fitting_set_variables_ui:
            self.grand_parent.fitting_set_variables_ui.update_table()

    def closeEvent(self, event=None):
        self.grand_parent.advanced_selection_ui = None

    def apply_button_clicked(self):
        QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        if self.ui.tabWidget.currentIndex() == 0:
            self.selection_table_selection_changed()
        else:
            self.lock_table_selection_changed()
        QApplication.restoreOverrideCursor()



#!/usr/bin/env python
"""
Linear bin
"""

import logging

import numpy as np

from ibeatles.tools.utilities import TimeSpectraKeys


class LinearBin:
    linear_bins = {
        TimeSpectraKeys.tof_array: None,
        TimeSpectraKeys.file_index_array: None,
        TimeSpectraKeys.lambda_array: None,
    }

    def __init__(self, parent=None, source_array=TimeSpectraKeys.file_index_array):
        self.parent = parent
        self.source_array = source_array
        self.logger = logging.getLogger("maverick")

    def create_linear_file_index_bin_array(self, bin_value=1):
        """creates the array of bins
        output will look like [[0,1],[2,3],[4,5]...]
        """
        original_array = np.array(self.parent.time_spectra[self.source_array])

        if self.source_array == TimeSpectraKeys.file_index_array:
            number_of_files = len(original_array)
            new_index_array = np.arange(0, number_of_files, bin_value)
            self.parent.full_bin_axis_requested = new_index_array

            linear_file_index_bin_array = [[] for _ in np.arange(len(new_index_array))]
            for _file_index, _bin in enumerate(original_array):
                result = np.where(_bin >= new_index_array)
                index = result[0][-1]
                linear_file_index_bin_array[index].append(_file_index)

            array_of_bins = linear_file_index_bin_array

        elif self.source_array == TimeSpectraKeys.tof_array:
            original_tof_array = np.array(self.parent.time_spectra[TimeSpectraKeys.tof_array])

            new_tof_array = np.arange(original_tof_array[0], original_tof_array[-1], bin_value)
            new_tof_array = np.append(new_tof_array, new_tof_array[-1] + bin_value)
            self.parent.full_bin_axis_requested = new_tof_array

            linear_tof_bin_array = [[] for _ in np.arange(len(new_tof_array) - 1)]
            for _tof_index, _bin in enumerate(original_tof_array):
                result = np.where(_bin >= new_tof_array)
                index = result[0][-1]
                linear_tof_bin_array[index].append(_tof_index)

            array_of_bins = linear_tof_bin_array

        elif self.source_array == TimeSpectraKeys.lambda_array:
            original_lambda_array = np.array(self.parent.time_spectra[TimeSpectraKeys.lambda_array])
            new_lambda_array = np.arange(original_lambda_array[0], original_lambda_array[-1], bin_value)
            new_lambda_array = np.append(new_lambda_array, original_lambda_array[-1] + bin_value)
            self.parent.full_bin_axis_requested = new_lambda_array

            linear_lambda_bin_array = [[] for _ in np.arange(len(new_lambda_array) - 1)]
            for _tof_index, _bin in enumerate(original_lambda_array):
                result = np.where(_bin >= new_lambda_array)
                index = result[0][-1]
                linear_lambda_bin_array[index].append(_tof_index)

            array_of_bins = linear_lambda_bin_array

        self.linear_bins[self.source_array] = array_of_bins
        self.logger.info(f"{self.source_array} array of bins: {array_of_bins}")

    def create_linear_lambda_array(self, lambda_value):
        """this method create the linear lambda array"""
        original_lambda_array = np.array(self.parent.time_spectra[TimeSpectraKeys.lambda_array])
        linear_bins = self._create_general_linear_array(stepping=lambda_value, original_array=original_lambda_array)
        self.linear_bins[TimeSpectraKeys.lambda_array] = linear_bins

    def _create_general_linear_array(self, stepping=None, original_array=None):
        """
        generic function used to create a linear bin array

        :param stepping: stepping bin value
        :param original_array: original array that will be used to determine when to stop
        :return: the linear bin array
        """
        left_value = original_array[0]
        right_value = stepping + left_value
        _linear_bins = []
        while right_value < original_array[-1]:
            _linear_bins.append(left_value)
            left_value = right_value
            right_value += stepping
        _linear_bins.append(right_value)
        return np.array(_linear_bins)

    def create_linear_bin_arrays(self):
        self.logger.info("Creating the other arrays")

        file_index_array_of_bins = self.linear_bins[self.source_array]

        original_tof_array = np.array(self.parent.time_spectra[TimeSpectraKeys.tof_array])
        original_lambda_array = np.array(self.parent.time_spectra[TimeSpectraKeys.lambda_array])

        linear_bins_tof_array = []
        linear_bins_lambda_array = []

        for _index, _bin in enumerate(file_index_array_of_bins):
            if _bin == []:
                linear_bins_tof_array.append([])
                linear_bins_lambda_array.append([])
                continue

            tof_bin = []
            lambda_bin = []
            for _file_index in _bin:
                tof_bin.append(original_tof_array[_file_index])
                lambda_bin.append(original_lambda_array[_file_index])
            linear_bins_tof_array.append(tof_bin)
            linear_bins_lambda_array.append(lambda_bin)

        self.linear_bins[TimeSpectraKeys.tof_array] = linear_bins_tof_array
        self.linear_bins[TimeSpectraKeys.lambda_array] = linear_bins_lambda_array
        self.linear_bins[TimeSpectraKeys.file_index_array] = file_index_array_of_bins

    def get_linear_file_index(self):
        return self.linear_bins[TimeSpectraKeys.file_index_array]

    def get_linear_tof(self):
        return self.linear_bins[TimeSpectraKeys.tof_array]

    def get_linear_lambda(self):
        return self.linear_bins[TimeSpectraKeys.lambda_array]



import numpy as np
from qtpy import QtCore, QtGui
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QSizePolicy,
    QSpacerItem,
    QTableWidgetItem,
    QTableWidgetSelectionRange,
    QWidget,
)

from ibeatles.utilities.widgets_handler import WidgetsHandler


class TableHandler:
    cell_str_format = "{:.3f}"
    cell_str_format = "{}"

    def __init__(self, table_ui=None):
        self.table_ui = table_ui

    def set_row_hidden(self, row=0, hide=True):
        self.table_ui.setRowHidden(row, hide)

    def select_everything(self, state):
        nbr_row = self.table_ui.rowCount()
        nbr_column = self.table_ui.columnCount()
        selection_range = QTableWidgetSelectionRange(0, 0, nbr_row - 1, nbr_column - 1)
        self.table_ui.setRangeSelected(selection_range, state)

    def row_count(self):
        return self.table_ui.rowCount()

    def column_count(self):
        return self.table_ui.columnCount()

    def select_rows(self, list_of_rows=None):
        self.table_ui.blockSignals(True)
        self.select_everything(False)
        nbr_column = self.table_ui.columnCount()

        for _row in list_of_rows:
            selection_range = QTableWidgetSelectionRange(_row, 0, _row, nbr_column - 1)
            self.table_ui.setRangeSelected(selection_range, True)
        self.table_ui.blockSignals(False)

    def add_this_row_to_selection(self, row=0):
        nbr_column = self.column_count()
        selection_range = QTableWidgetSelectionRange(row, 0, row, nbr_column - 1)
        self.table_ui.setRangeSelected(selection_range, True)

    def remove_all_rows(self):
        self.table_ui.blockSignals(True)
        nbr_row = self.table_ui.rowCount()
        for _ in np.arange(nbr_row):
            self.remove_row(row=0)
        self.table_ui.blockSignals(False)

    def remove_row(self, row=-1):
        self.table_ui.removeRow(row)

    def remove_all_columns(self):
        self.table_ui.blockSignals(True)
        nbr_column = self.table_ui.columnCount()
        for _ in np.arange(nbr_column):
            self.remove_column(column=0)
        self.table_ui.blockSignals(False)

    def remove_column(self, column=-1):
        self.table_ui.removeColumn(column)

    def full_reset(self):
        self.remove_all_rows()
        self.remove_all_columns()

    def get_rows_of_table_selected(self):
        if self.table_ui is None:
            return []

        selected_ranges = self.table_ui.selectedRanges()
        if len(selected_ranges) == 0:
            return []

        list_row_selected = []
        for _selection in selected_ranges:
            top_row = _selection.topRow()
            bottom_row = _selection.bottomRow()
            if top_row == bottom_row:
                list_row_selected.append(top_row)
            else:
                _range = np.arange(top_row, bottom_row + 1)
                for _row in _range:
                    list_row_selected.append(_row)

        list_row_selected.sort()
        return list_row_selected

    def get_row_selected(self) -> int:
        if self.table_ui is None:
            return -1
        list_selection = self.table_ui.selectedRanges()
        try:
            first_selection = list_selection[0]
        except IndexError:
            return -1
        return first_selection.topRow()

    def get_column_selected(self):
        if self.table_ui is None:
            return -1
        list_selection = self.table_ui.selectedRanges()
        try:
            first_selection = list_selection[0]
        except IndexError:
            return -1
        return first_selection.leftColumn()

    def get_cell_selected(self):
        list_selection = self.table_ui.selectedRanges()
        first_selection = list_selection[0]
        row = first_selection.topRow()
        col = first_selection.leftColumn()
        return row, col

    def get_selection(self):
        return self.table_ui.selectedRanges()

    def get_item_str_from_cell(self, row=-1, column=-1):
        if self.table_ui.item(row, column):
            return str(self.table_ui.item(row, column).text())
        else:
            return None

    def get_item_float_from_cell(self, row=-1, column=-1):
        item_selected = self.table_ui.item(row, column).text()
        try:
            float_item = float(item_selected)
        except ValueError:
            return np.nan
        return float_item

    def select_cell(self, row=0, column=0):
        self.select_everything(False)
        range_selected = QTableWidgetSelectionRange(row, column, row, column)
        self.table_ui.setRangeSelected(range_selected, True)

    def select_row(self, row=0):
        if row < 0:
            row = 0
        self.table_ui.selectRow(row)

    def set_column_names(self, column_names=None):
        self.table_ui.setHorizontalHeaderLabels(column_names)

    def get_column_names(self):
        nbr_column = self.column_count()
        names = []
        for _column in np.arange(nbr_column):
            names.append(self.table_ui.horizontalHeaderItem(_column).text())
        return names

    def set_row_names(self, row_names=None):
        self.table_ui.setVerticalHeaderLabels(row_names)

    def set_column_sizes(self, column_sizes=None):
        for _col, _size in enumerate(column_sizes):
            self.table_ui.setColumnWidth(_col, _size)

    def set_column_width(self, column_width=None):
        self.set_column_sizes(column_sizes=column_width)

    def set_row_height(self, row_height=None):
        for _row, _height in enumerate(row_height):
            self.table_ui.setRowHeight(_row, _height)

    def insert_empty_row(self, row=0):
        self.table_ui.insertRow(row)

    def insert_row(self, row=0, list_col_name=None):
        """row is the row number"""
        self.table_ui.insertRow(row)
        for column, _text in enumerate(list_col_name):
            _item = QTableWidgetItem(_text)
            self.table_ui.setItem(row, column, _item)

    def is_item(self, row=0, column=0):
        if self.table_ui.item(row, column):
            return True
        return False

    def get_widget(self, row=-1, column=-1):
        _widget = self.table_ui.cellWidget(row, column)
        return _widget

    def insert_column(self, column):
        self.table_ui.insertColumn(column)

    def insert_empty_column(self, column):
        self.insert_column(column)

    def set_item_with_str(self, row=0, column=0, cell_str=""):
        self.table_ui.item(row, column).setText(cell_str)

    def set_item_with_float(self, row=0, column=0, float_value=""):
        if (str(float_value) == "None") or (str(float_value) == "N/A"):
            _str_value = "N/A"
        else:
            _str_value = self.cell_str_format.format(float(float_value))
        self.table_ui.item(row, column).setText(_str_value)

    def insert_item_with_float(self, row=0, column=0, float_value="", format_str="{}"):
        if (str(float_value) == "None") or (str(float_value) == "N/A"):
            _str_value = "N/A"
        else:
            _str_value = format_str.format(float(float_value))
        _item = QTableWidgetItem(_str_value)
        self.table_ui.setItem(row, column, _item)

    def insert_item(
        self,
        row=0,
        column=0,
        value="",
        format_str="{}",
        editable=True,
        align_center=False,
    ):
        if type(value) is str:
            _str_value = value
        elif value is None:
            _str_value = str(np.nan)
        else:
            _str_value = format_str.format(value)

        _item = QTableWidgetItem(_str_value)

        if not editable:
            _item.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
        if align_center:
            _item.setTextAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)

        self.table_ui.setItem(row, column, _item)

    def insert_widget(self, row=0, column=0, widget=None, centered=False):
        if not centered:
            self.table_ui.setCellWidget(row, column, widget)
        else:
            spacer1 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
            spacer2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
            line_widget = QWidget()
            hori_layout = QHBoxLayout()
            hori_layout.addItem(spacer1)
            hori_layout.addWidget(widget)
            hori_layout.addItem(spacer2)
            line_widget.setLayout(hori_layout)
            self.table_ui.setCellWidget(row, column, line_widget)

    def set_background_color(self, row=0, column=0, qcolor=QtGui.QColor(0, 255, 255)):
        _item = self.table_ui.item(row, column)
        _item.setBackground(qcolor)

    def set_background_color_of_row(self, row=0, qcolor=QtGui.QColor(0, 255, 255)):
        nbr_column = self.column_count()
        for _col in np.arange(nbr_column):
            self.set_background_color(row=row, column=_col, qcolor=qcolor)

    def fill_table_with(self, list_items=None, editable_columns_boolean=None, block_signal=False):
        """
        :param:
        list_items: 2D array of text to put in the table
            ex: list_items = [ ['file1', 10, 20'], ['file2', 20, 30] ...]
        editable_columns_boolean: which columns are editable
            ex: editable_columns_boolean = [False, True, True]
        block_signals: block or not any signal emitted by the table
        """
        if block_signal:
            WidgetsHandler.block_signals(ui=self.table_ui, status=True)

        self.remove_all_rows()

        for _row_index, _row_entry in enumerate(list_items):
            self.insert_empty_row(_row_index)
            for _column_index, _text in enumerate(list_items[_row_index]):
                if _row_index == 0:
                    editable_flag = False
                else:
                    editable_flag = editable_columns_boolean[_column_index]
                self.insert_item(
                    row=_row_index,
                    column=_column_index,
                    value=_text,
                    editable=editable_flag,
                )

        if block_signal:
            WidgetsHandler.block_signals(ui=self.table_ui, status=False)

    def block_signals(self, state=True):
        self.table_ui.blockSignals(state)
        QApplication.processEvents()



import logging

import pandas as pd
from qtpy.QtWidgets import QFileDialog


class Export:
    header = [
        "x0",
        "y0",
        "x1",
        "y1",
        "row_index",
        "column_index",
        "lock",
        "active",
        "fitting_confidence",
        "d_spacing_value",
        "d_spacing_err",
        "sigma_value",
        "sigma_err",
        "intensity_value",
        "intensity_err",
        "alpha_value",
        "alpha_err",
        "a1_value",
        "a1_err",
        "a2_value",
        "a2_err",
        "a5_value",
        "a5_err",
        "a6_value",
        "a6_err",
    ]

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def run(self):
        logging.info("Exporting table")
        default_file_name = str(self.grand_parent.ui.normalized_folder.text()) + "_fitting_table.csv"
        table_file = QFileDialog.getSaveFileName(
            self.grand_parent,
            "Select or Define Name of File!",
            default_file_name,
            "CSV (*.csv)",
        )

        if table_file[0]:
            table_file = table_file[0]
            logging.info(f" table file selected: {table_file}")
            table_dictionary = self.grand_parent.march_table_dictionary
            o_table_formatted = FormatTableForExport(table=table_dictionary)
            pandas_data_frame = o_table_formatted.pandas_data_frame
            header = self.header
            pandas_data_frame.to_csv(table_file, header=header)
            logging.info(f" Table has been exporting in {table_file}")
        else:
            logging.info(" User canceled exporting table!")


class FormatTableForExport(object):
    pandas_data_frame = []

    def __init__(self, table={}):
        pandas_table = []

        for _key in table:
            _entry = table[_key]

            x0 = _entry["bin_coordinates"]["x0"]
            y0 = _entry["bin_coordinates"]["y0"]
            x1 = _entry["bin_coordinates"]["x1"]
            y1 = _entry["bin_coordinates"]["y1"]

            row_index = _entry["row_index"]
            column_index = _entry["column_index"]

            lock = _entry["lock"]
            active = _entry["active"]

            fitting_confidence = _entry["fitting_confidence"]

            [d_spacing_val, d_spacing_err] = FormatTableForExport.get_val_err_fixed(_entry["d_spacing"])

            [sigma_val, sigma_err] = FormatTableForExport.get_val_err_fixed(_entry["sigma"])

            [intensity_val, intensity_err] = FormatTableForExport.get_val_err_fixed(_entry["intensity"])

            [alpha_val, alpha_err] = FormatTableForExport.get_val_err_fixed(_entry["alpha"])

            [a1_val, a1_err] = FormatTableForExport.get_val_err_fixed(_entry["a1"])

            [a2_val, a2_err] = FormatTableForExport.get_val_err_fixed(_entry["a2"])

            [a5_val, a5_err] = FormatTableForExport.get_val_err_fixed(_entry["a5"])

            [a6_val, a6_err] = FormatTableForExport.get_val_err_fixed(_entry["a6"])

            _row = [
                x0,
                x1,
                y0,
                y1,
                row_index,
                column_index,
                lock,
                active,
                fitting_confidence,
                d_spacing_val,
                d_spacing_err,
                sigma_val,
                sigma_err,
                intensity_val,
                intensity_err,
                alpha_val,
                alpha_err,
                a1_val,
                a1_err,
                a2_val,
                a2_err,
                a5_val,
                a5_err,
                a6_val,
                a6_err,
            ]

            pandas_table.append(_row)

        pandas_data_frame = pd.DataFrame.from_dict(pandas_table)
        self.pandas_data_frame = pandas_data_frame

    @staticmethod
    def get_val_err_fixed(item):
        return [item["val"], item["err"]]



#!/usr/bin/env python
"""
Initialization
"""

import pyqtgraph as pg
from qtpy.QtWidgets import QProgressBar, QVBoxLayout

from ibeatles import interact_me_style


class Initialization:
    def __init__(self, parent=None):
        self.parent = parent

    def all(self):
        self.widgets()
        self.pyqtgraph()

    def widgets(self):
        # progress bar
        self.parent.eventProgress = QProgressBar(self.parent.ui.statusbar)
        self.parent.eventProgress.setMinimumSize(20, 14)
        self.parent.eventProgress.setMaximumSize(540, 100)
        self.parent.eventProgress.setVisible(False)
        self.parent.ui.statusbar.addPermanentWidget(self.parent.eventProgress)

        # show what to do next
        self.parent.ui.select_folder_pushButton.setStyleSheet(interact_me_style)

    def pyqtgraph(self):
        self.parent.ui.image_view = pg.ImageView(view=pg.PlotItem())
        self.parent.ui.image_view.ui.roiBtn.hide()
        self.parent.ui.image_view.ui.menuBtn.hide()

        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(self.parent.ui.image_view)
        self.parent.ui.widget.setLayout(vertical_layout)
        self.parent.ui.line_view = None



#!/usr/bin/env python
"""
Combine module
"""

import numpy as np
import pyqtgraph as pg

from ibeatles.tools.tof_combine import SessionKeys as TofCombineSessionKeys
from ibeatles.tools.tof_combine.utilities.get import Get
from ibeatles.tools.utilities import CombineAlgorithm


class Combine:
    def __init__(self, parent=None):
        self.parent = parent

    def run(self):
        o_get = Get(parent=self.parent)
        combine_algorithm = o_get.combine_algorithm()
        list_array_to_combine = o_get.list_array_to_combine()

        if list_array_to_combine is None:
            self.parent.combine_image_view.clear()
            self.parent.combine_image_view.removeItem(self.parent.combine_roi_item_id)
            self.parent.combine_data = None

        else:
            # combine using algorithm defined
            [nbr_folder_to_combine, nbr_files, width, height] = np.shape(list_array_to_combine)

            if nbr_folder_to_combine > 1:
                if combine_algorithm == CombineAlgorithm.mean:
                    combine_arrays = np.mean(list_array_to_combine, axis=0)
                    integrated_arrays = np.transpose(np.mean(combine_arrays, axis=0))

                elif combine_algorithm == CombineAlgorithm.median:
                    combine_arrays = np.median(list_array_to_combine, axis=0)
                    integrated_arrays = np.transpose(np.median(combine_arrays, axis=0))

                else:
                    raise NotImplementedError("Algorithm not implemented!")
            else:
                combine_arrays = list_array_to_combine[0]
                integrated_arrays = np.transpose(np.mean(combine_arrays, axis=0))

            combine_arrays = np.squeeze(combine_arrays)
            self.parent.combine_data = combine_arrays
            self.parent.live_combine_image = integrated_arrays

            # display integrated
            if self.parent.visualize_flag:
                self.parent.combine_image_view.setImage(integrated_arrays)

                # initialize ROI if first time, otherwise use same region
                roi_dict = self.parent.session[TofCombineSessionKeys.combine_roi]
                x0 = roi_dict["x0"]
                y0 = roi_dict["y0"]
                width = roi_dict["width"]
                height = roi_dict["height"]
                if self.parent.combine_roi_item_id:
                    self.parent.combine_image_view.removeItem(self.parent.combine_roi_item_id)

                roi_item = pg.ROI([x0, y0], [width, height])
                roi_item.addScaleHandle([1, 1], [0, 0])
                self.parent.combine_image_view.addItem(roi_item)
                roi_item.sigRegionChanged.connect(self.parent.combine_roi_changed)
                self.parent.combine_roi_item_id = roi_item



#!/usr/bin/env python
"""
List of h, k, l, lambda and d0
"""

import numpy as np
from qtpy.QtWidgets import QDialog

from ibeatles import Material, load_ui
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.table_handler import TableHandler


class ListHKLLambdaD0Handler:
    def __init__(self, parent=None):
        if parent.list_hkl_lambda_d0_ui is None:
            list_dialog = ListHKLLambdaD0(parent=parent)
            list_dialog.show()
            parent.list_hkl_lambda_d0_ui = list_dialog
        else:
            parent.list_hkl_lambda_d0_ui.setFocus()
            parent.list_hkl_lambda_d0_ui.activateWindow()


class ListHKLLambdaD0(QDialog):
    new_element: dict = {}

    def __init__(self, parent=None):
        self.parent = parent

        QDialog.__init__(self, parent=parent)
        self.ui = load_ui("ui_list_hkl.ui", baseinstance=self)
        self.setWindowTitle("List of h, k, l, lambda and d0")

        self.init_table()
        self.populate_table()

    def init_table(self):
        o_table = TableHandler(table_ui=self.ui.tableWidget)
        column_sizes = [70, 70, 70, 70, 70]
        o_table.set_column_sizes(column_sizes=column_sizes)

        column_names = ["h", "k", "l", "\u03bb", "d\u2090"]
        o_table.set_column_names(column_names=column_names)

    def populate_table(self):
        """
        Take the self.parent.selected_element_bragg_edges_array, self.parent.selected_element_hkl_array
        and the current selected element to populate the corresponding infos
        """
        o_gui = GuiHandler(parent=self.parent)
        element_name = str(o_gui.get_text_selected(ui=self.parent.ui.list_of_elements))
        self.ui.selected_element_value.setText(element_name)

        if element_name in self.parent.user_defined_bragg_edge_list.keys():
            user_defined_bragg_edge_list = self.parent.user_defined_bragg_edge_list[element_name]
            list_hkl_d0 = user_defined_bragg_edge_list[Material.hkl_d0]
            o_table = TableHandler(table_ui=self.ui.tableWidget)
            _row = 0
            for _key in list_hkl_d0.keys():
                _entry = list_hkl_d0[_key]
                if _entry["h"] is None:
                    continue
                o_table.insert_empty_row(row=_row)

                _h = _entry["h"]
                _k = _entry["k"]
                _l = _entry["l"]
                _d0 = _entry["d0"]
                _lambda = 2 * float(_d0)

                o_table.insert_item(row=_row, column=0, value=_h)

                o_table.insert_item(row=_row, column=1, value=_k)

                o_table.insert_item(row=_row, column=2, value=_l)

                o_table.insert_item(row=_row, column=3, value=_lambda, format_str="{:.3f}")

                o_table.insert_item(row=_row, column=4, value=_d0, format_str="{:.3f}")

                _row += 1

        else:  # element found in the default list
            selected_element_bragg_edges_array = self.parent.selected_element_bragg_edges_array
            selected_element_hkl_array = self.parent.selected_element_hkl_array

            nbr_row = len(selected_element_hkl_array)
            o_table = TableHandler(table_ui=self.ui.tableWidget)
            for _row in np.arange(nbr_row):
                o_table.insert_empty_row(row=_row)
                [_h, _k, _l] = selected_element_hkl_array[_row]
                _lambda = selected_element_bragg_edges_array[_row]

                o_table.insert_item(row=_row, column=0, value=_h)

                o_table.insert_item(row=_row, column=1, value=_k)

                o_table.insert_item(row=_row, column=2, value=_l)

                o_table.insert_item(row=_row, column=3, value=_lambda, format_str="{:.3f}")

                o_table.insert_item(row=_row, column=4, value=_lambda / 2.0, format_str="{:.3f}")

    def closeEvent(self, ev):
        self.parent.list_hkl_lambda_d0_ui = None

    def close_clicked(self):
        self.parent.list_hkl_lambda_d0_ui = None
        self.close()

    def refresh_populate_table(self):
        o_table = TableHandler(table_ui=self.ui.tableWidget)
        o_table.remove_all_rows()
        self.populate_table()



#!/usr/bin/env python
"""
Time spectra module
"""

from ibeatles.tools.utilities import TimeSpectraKeys


def format_str(input_list, format_str="{}", factor=1, data_type=TimeSpectraKeys.file_index_array):
    """
    format the list of file_index, tof or lambda to fill the manual bin table
    :param input_list:
    :param format_str:
    :param factor:
    :param data_type:
    :return:
    """
    if data_type == TimeSpectraKeys.file_index_array:
        if len(input_list) == 1:
            return format_str.format(input_list[0] * factor)
        elif len(input_list) == 2:
            return format_str.format(input_list[0] * factor) + ", " + format_str.format(input_list[1] * factor)
        else:
            return format_str.format(input_list[0] * factor) + " - " + format_str.format(input_list[-1] * factor)
    else:
        if len(input_list) == 1:
            return format_str.format(input_list[0] * factor)
        else:
            return format_str.format(input_list[0] * factor) + " - " + format_str.format(input_list[-1] * factor)
